from typing import Dict, Optional

import pytest
from karp.foundation.value_objects.unique_id import make_unique_id

from karp.lex.domain import errors

from karp.lex.application import repositories
from karp.lex.application.repositories import (
    ResourceUnitOfWork,
    EntryUowRepositoryUnitOfWork,
)
from karp.lex.domain import commands

from . import adapters, factories

# from karp.domain import errors, events
# from karp.domain.value_objects.unique_id import make_unique_id
# from karp.lex.domain import commands
# from karp.services import unit_of_work

# from karp.tests.unit.adapters import (InMemoryEntryUowFactory, InMemoryResourceUnitOfWork,
#                                       bootstrap_test_app)


class TestAddEntry:
    def test_cannot_add_entry_to_nonexistent_resource(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        with pytest.raises(errors.ResourceNotFound):
            lex_ctx.command_bus.dispatch(
                factories.AddEntryFactory(  # type: ignore [arg-type]
                    resource_id="non_existent",
                )
            )

    def test_add_entry(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {"baseform": {"type": "string", "required": True}},
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        cmd3 = factories.AddEntryFactory(
            resource_id=cmd2.resource_id,
            entry={"baseform": entry_id},
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        entry_uow_repo_uow = lex_ctx.container.get(EntryUowRepositoryUnitOfWork)  # type: ignore [misc]
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)  # type: ignore [arg-type]
        entry = uow.repo.by_id(cmd3.entity_id)  # type: ignore [arg-type]
        assert entry is not None
        assert entry.id == cmd3.entity_id
        # assert entry.entry_id == entry_id
        # assert entry.repo_id == cmd3.resource_id

        assert entry.body == {"baseform": entry_id}
        assert entry.last_modified_by == cmd3.user

        assert uow.was_committed  # type: ignore [attr-defined]

        # assert (
        #     bus.ctx.index_uow.repo.indicies[resource_id].entries[entry_id].id
        #     == entry_id
        # )
        # assert bus.ctx.index_uow.was_committed #type: ignore [attr-defined]

    @pytest.mark.skip(reason="we don't use entry_id anymore")
    def test_create_entry_with_same_entry_id_raises(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {"baseform": {"type": "string", "required": True}},
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        cmd3 = factories.AddEntryFactory(
            resource_id=cmd2.resource_id,
            entry={"baseform": entry_id},
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        with pytest.raises(errors.IntegrityError):
            lex_ctx.command_bus.dispatch(
                factories.AddEntryFactory(  # type: ignore [arg-type]
                    resource_id=cmd3.resource_id,
                    entry={"baseform": entry_id},
                ),
            )


class TestUpdateEntry:
    def test_update_entry(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {
                    "baseform": {"type": "string", "required": True},
                    "a": {
                        "type": "string",
                    },
                },
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        cmd3 = factories.AddEntryFactory(
            resource_id=cmd2.resource_id,
            entry={"baseform": entry_id, "a": "orig"},
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        lex_ctx.command_bus.dispatch(
            factories.UpdateEntryFactory(  # type: ignore [arg-type]
                entity_id=cmd3.entity_id,
                # entry_id=entry_id,
                version=1,
                resource_id=cmd3.resource_id,
                entry={"baseform": entry_id, "a": "changed", "b": "added"},
            ),
        )

        entry_uow_repo_uow = lex_ctx.container.get(EntryUowRepositoryUnitOfWork)  # type: ignore [misc]
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)  # type: ignore [arg-type]
        assert uow.was_committed  # type: ignore [attr-defined]

        entry = uow.repo.by_id(cmd3.entity_id)  # type: ignore [arg-type]
        assert entry is not None
        assert entry.body["a"] == "changed"
        assert entry.body["b"] == "added"
        assert entry.version == 2
        # assert (
        #     bus.ctx.index_uow.repo.indicies[resource_id]
        #     .entries[entry.entry_id]
        #     .entry["_entry_version"]
        #     == entry.version
        # )
        # assert bus.ctx.index_uow.was_committed #type: ignore [attr-defined]

    def test_cannot_update_entry_in_nonexistent_resource(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        with pytest.raises(errors.ResourceNotFound):
            lex_ctx.command_bus.dispatch(
                commands.UpdateEntry(
                    entity_id=make_unique_id(),
                    resource_id="non_existent",
                    # entry_id="a",
                    version=3,
                    entry={},
                    user="kristoff@example.com",
                    message="update",
                )
            )

    def test_update_of_entry_id_changes_entry_id(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {
                    "baseform": {"type": "string", "required": True},
                },
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        cmd3 = factories.AddEntryFactory(
            resource_id=cmd2.resource_id,
            entry={"baseform": entry_id},
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        new_entry_id = "gamma"
        lex_ctx.command_bus.dispatch(
            factories.UpdateEntryFactory(  # type: ignore [arg-type]
                entity_id=cmd3.entity_id,
                # entry_id=entry_id,
                version=1,
                resource_id=cmd3.resource_id,
                entry={"baseform": new_entry_id},
            ),
        )

        entry_uow_repo_uow = lex_ctx.container.get(EntryUowRepositoryUnitOfWork)  # type: ignore [misc]
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)  # type: ignore [arg-type]
        assert uow.was_committed  # type: ignore [attr-defined]

        entry = uow.repo.by_id(cmd3.entity_id)  # type: ignore [arg-type]
        # assert entry.entry_id == new_entry_id
        assert entry.version == 2

        # entry = uow.repo.by_entry_id(new_entry_id)
        # assert entry.id == cmd3.entity_id

        # with pytest.raises(errors.EntryNotFound):
        #     uow.repo.by_entry_id(entry_id)


class TestDeleteEntry:
    def test_can_delete_entry(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {
                    "baseform": {"type": "string", "required": True},
                    "a": {
                        "type": "string",
                    },
                },
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        cmd3 = factories.AddEntryFactory(
            resource_id=cmd2.resource_id,
            entry={"baseform": entry_id, "a": "orig"},
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        lex_ctx.command_bus.dispatch(
            commands.DeleteEntry(
                entity_id=cmd3.entity_id,
                # TODO: should we give version to delete entry for sync
                # version=1,
                resource_id=cmd3.resource_id,
                message="deleted",
                user="bob",
            ),
        )

        entry_uow_repo_uow = lex_ctx.container.get(EntryUowRepositoryUnitOfWork)  # type: ignore [misc]
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)  # type: ignore [arg-type]
        assert uow.was_committed  # type: ignore [attr-defined]

        entry = uow.repo.by_id(cmd3.entity_id)  # type: ignore [arg-type]
        assert entry.version == 2
        assert entry.discarded

        # entry = uow.repo._by_entry_id(entry_id)
        # assert entry.version == 2
        # assert entry.discarded
        # assert (
        #     entry.entry_id not in bus.ctx.index_uow.repo.indicies[resource_id].entries
        # )
        # assert bus.ctx.index_uow.was_committed # type: ignore [attr-defined]


class TestAddEntries:
    def test_cannot_add_entries_to_nonexistent_resource(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        with pytest.raises(errors.ResourceNotFound):
            lex_ctx.command_bus.dispatch(
                factories.AddEntriesFactory(
                    resource_id="non_existent",  # type: ignore [arg-type]
                )
            )

    def test_add_entry(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {"baseform": {"type": "string", "required": True}},
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        cmd3 = factories.AddEntriesFactory(
            resource_id=cmd2.resource_id,
            entries=[{"baseform": entry_id}],
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        entry_uow_repo_uow = lex_ctx.container.get(EntryUowRepositoryUnitOfWork)  # type: ignore [misc]
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)  # type: ignore [arg-type]
        # entry = uow.repo.by_id(cmd3.entity_id)
        # assert entry is not None
        # assert entry.entry_id == entry_id
        # # assert entry.repo_id == cmd3.resource_id

        # assert entry.body == {"baseform": entry_id}
        # assert entry.last_modified_by == cmd3.user

        assert uow.was_committed  # type: ignore [attr-defined]

        # assert (
        #     bus.ctx.index_uow.repo.indicies[resource_id].entries[entry_id].id
        #     == entry_id
        # )
        # assert bus.ctx.index_uow.was_committed #type: ignore [attr-defined]

    @pytest.mark.skip(reason="we don't use entry_id")
    def test_add_entry_with_same_entry_id_raises(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {"baseform": {"type": "string", "required": True}},
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        cmd3 = factories.AddEntriesFactory(
            resource_id=cmd2.resource_id,
            entries=[{"baseform": entry_id}],
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        with pytest.raises(errors.IntegrityError):
            lex_ctx.command_bus.dispatch(
                factories.AddEntriesFactory(  # type: ignore [arg-type]
                    resource_id=cmd3.resource_id,
                    entries=[{"baseform": entry_id}],
                ),
            )


class TestAddEntriesInChunks:
    def test_cannot_add_entries_to_nonexistent_resource(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        with pytest.raises(errors.ResourceNotFound):
            lex_ctx.command_bus.dispatch(
                factories.AddEntriesInChunksFactory(  # type: ignore [arg-type]
                    resource_id="non_existent",
                )
            )

    def test_add_entry(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {"baseform": {"type": "string", "required": True}},
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        entity_id = make_unique_id()
        cmd3 = factories.AddEntriesInChunksFactory(
            resource_id=cmd2.resource_id,
            entries=[{"baseform": entry_id}],
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        entry_uow_repo_uow = lex_ctx.container.get(EntryUowRepositoryUnitOfWork)  # type: ignore [misc]
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)  # type: ignore [arg-type]
        # entry = uow.repo.by_id(entry_id)
        # assert entry is not None
        # assert entry.entry_id == entry_id
        # # assert entry.repo_id == cmd3.resource_id

        # assert entry.body == {"baseform": entry_id}
        # assert entry.last_modified_by == cmd3.user

        assert uow.was_committed  # type: ignore [attr-defined]

        # assert (
        #     bus.ctx.index_uow.repo.indicies[resource_id].entries[entry_id].id
        #     == entry_id
        # )
        # assert bus.ctx.index_uow.was_committed #type: ignore [attr-defined]

    @pytest.mark.skip(reason="we don't use entry_id")
    def test_add_entry_with_same_entry_id_raises(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {"baseform": {"type": "string", "required": True}},
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        cmd3 = factories.AddEntriesInChunksFactory(
            resource_id=cmd2.resource_id,
            entries=[{"baseform": entry_id}],
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        with pytest.raises(errors.IntegrityError):
            lex_ctx.command_bus.dispatch(
                factories.AddEntriesInChunksFactory(  # type: ignore [arg-type]
                    resource_id=cmd3.resource_id,
                    entries=[{"baseform": entry_id}],
                ),
            )


class TestImportEntries:
    def test_cannot_import_entries_to_nonexistent_resource(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        with pytest.raises(errors.ResourceNotFound):
            lex_ctx.command_bus.dispatch(
                factories.ImportEntriesFactory(  # type: ignore [arg-type]
                    resource_id="non_existent",
                )
            )

    def test_import_entries(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {"baseform": {"type": "string", "required": True}},
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        entity_id = make_unique_id()
        cmd3 = factories.ImportEntriesFactory(
            resource_id=cmd2.resource_id,
            entries=[
                {
                    "entity_id": entity_id,
                    "entry": {"baseform": entry_id},
                    "user": "user1",
                }
            ],
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        entry_uow_repo_uow = lex_ctx.container.get(EntryUowRepositoryUnitOfWork)  # type: ignore [misc]
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)  # type: ignore [arg-type]
        entry = uow.repo.by_id(entity_id)
        assert entry is not None
        # assert entry.entry_id == entry_id
        # assert entry.repo_id == cmd3.resource_id

        assert entry.body == {"baseform": entry_id}
        assert entry.last_modified_by == "user1"

        assert uow.was_committed  # type: ignore [attr-defined]

        # assert (
        #     bus.ctx.index_uow.repo.indicies[resource_id].entries[entry_id].id
        #     == entry_id
        # )
        # assert bus.ctx.index_uow.was_committed #type: ignore [attr-defined]

    @pytest.mark.skip(reason="we don't use entry_id")
    def test_import_entries_with_same_entry_id_raises(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {"baseform": {"type": "string", "required": True}},
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        cmd3 = factories.ImportEntriesFactory(
            resource_id=cmd2.resource_id,
            entries=[
                {
                    "entry": {"baseform": entry_id},
                }
            ],
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        with pytest.raises(errors.IntegrityError):
            lex_ctx.command_bus.dispatch(
                factories.ImportEntriesFactory(  # type: ignore [arg-type]
                    resource_id=cmd3.resource_id,
                    entries=[{"entry": {"baseform": entry_id}}],
                ),
            )


class TestImportEntriesInChunks:
    def test_cannot_import_entries_to_nonexistent_resource(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        with pytest.raises(errors.ResourceNotFound):
            lex_ctx.command_bus.dispatch(  # type: ignore [arg-type]
                factories.ImportEntriesInChunksFactory(  # type: ignore [arg-type]
                    resource_id="non_existent",
                )
            )

    def test_import_entries(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {"baseform": {"type": "string", "required": True}},
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        entity_id = make_unique_id()
        cmd3 = factories.ImportEntriesInChunksFactory(
            resource_id=cmd2.resource_id,
            entries=[
                {
                    "entity_id": entity_id,
                    "entry": {"baseform": entry_id},
                    "user": "user1",
                }
            ],
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        entry_uow_repo_uow = lex_ctx.container.get(EntryUowRepositoryUnitOfWork)  # type: ignore [misc]
        uow = entry_uow_repo_uow.repo.get_by_id(cmd2.entry_repo_id)  # type: ignore [arg-type]
        entry = uow.repo.by_id(entity_id)  # type: ignore [attr.defined]
        assert entry is not None
        # assert entry.entry_id == entry_id
        # assert entry.repo_id == cmd3.resource_id

        assert entry.body == {"baseform": entry_id}
        assert entry.last_modified_by == "user1"

        assert uow.was_committed  # type: ignore [attr-defined]

        # assert (
        #     bus.ctx.index_uow.repo.indicies[resource_id].entries[entry_id].id
        #     == entry_id
        # )
        # assert bus.ctx.index_uow.was_committed #type: ignore [attr-defined]

    @pytest.mark.skip(reason="we don't use entry_id")
    def test_import_entries_with_same_entry_id_raises(
        self,
        lex_ctx: adapters.UnitTestContext,
    ):
        cmd1 = factories.CreateEntryRepositoryFactory()
        lex_ctx.command_bus.dispatch(cmd1)  # type: ignore [arg-type]

        cmd2 = factories.CreateResourceFactory(
            entry_repo_id=cmd1.entity_id,
            config={
                "sort": ["baseform"],
                "fields": {"baseform": {"type": "string", "required": True}},
                "id": "baseform",
            },
        )
        lex_ctx.command_bus.dispatch(cmd2)  # type: ignore [arg-type]

        entry_id = "beta"
        cmd3 = factories.ImportEntriesInChunksFactory(
            resource_id=cmd2.resource_id,
            entries=[
                {
                    "entry": {"baseform": entry_id},
                }
            ],
        )

        lex_ctx.command_bus.dispatch(cmd3)  # type: ignore [arg-type]

        with pytest.raises(errors.IntegrityError):
            lex_ctx.command_bus.dispatch(
                factories.ImportEntriesInChunksFactory(  # type: ignore [arg-type]
                    resource_id=cmd3.resource_id,
                    entries=[{"entry": {"baseform": entry_id}}],
                ),
            )



import typing

from karp.domain import commands
from karp.domain.value_objects.unique_id import make_unique_id


def make_create_resource_command(
    resource_id: str, config: typing.Optional[typing.Dict] = None
) -> commands.CreateResource:
    config = config or {
        "fields": {},
        "id": "id",
    }
    return commands.CreateResource(
        id=make_unique_id(),
        resource_id=resource_id,
        name=resource_id.upper(),
        config=config,
        message="create resource",
        created_by="kristoff@example.com",
    )



from fastapi import APIRouter

from karp.webapp.routes import (
    entries_api,
    health_api,
    history_api,
    query_api,
    resources_api,
    stats_api,
)


router = APIRouter()


router.include_router(
    entries_api.router,
    prefix="/entries",
    tags=["Entries"],
)
router.include_router(health_api.router, prefix="/healthz")
router.include_router(
    history_api.router,
    prefix="/history",
    tags=["Entries", "History"],
)
router.include_router(
    query_api.router,
    prefix="/query",
    tags=["Querying"],
)
router.include_router(
    resources_api.router,
    prefix="/resources",
    tags=["Resources"],
)
router.include_router(
    stats_api.router,
    prefix="/stats",
    tags=["Statistics"],
)



from typing import Dict

import pytest

from karp.lex.domain import errors, events
from karp.lex.domain import entities
from karp.foundation.value_objects import unique_id
from karp.tests import common_data

from karp.tests.unit.lex import factories


def random_entry(entry_id: str = None, body: Dict = None) -> entities.Entry:
    return entities.create_entry(
        entity_id=unique_id.make_unique_id(),
        repo_id=unique_id.make_unique_id(),
        # entry_id=entry_id or "a",
        body=body or {},
        message="add",
        last_modified_by="kristoff@example.com",
        last_modified=12345.67,
    )


def test_new_entry_has_event():
    entry = random_entry()
    assert entry.domain_events[-1] == events.EntryAdded(
        entity_id=entry.id,
        repo_id=entry.repo_id,
        body=entry.body,
        user=entry.last_modified_by,
        timestamp=entry.last_modified,
        message=entry.message,
    )


def test_discarded_entry_has_event():
    entry = random_entry()
    entry.discard(
        user="alice@example.org",
        message="bad",
        timestamp=123.45,
    )
    assert entry.discarded
    assert entry.domain_events[-1] == events.EntryDeleted(
        entity_id=entry.id,
        # entry_id=entry.entry_id,
        repo_id=entry.repo_id,
        user=entry.last_modified_by,
        timestamp=entry.last_modified,
        message=entry.message,
        version=2,
    )


# @pytest.mark.parametrize(
#     "field,value",
#     [
#         ("entry_id", "new..1"),
#         # ("body", {"b": "r"}),
#         # ("status", entities.EntryStatus.IN_REVIEW),
#     ],
# )
# def test_entry_update_updates(field, value):
#     entry = random_entry(entry_id="test..2", body={"a": ["1", "e"]})

#     previous_last_modified = entry.last_modified
#     previous_last_modified_by = entry.last_modified_by
#     message = f"Updated {field}"

#     user = "Test User"

#     setattr(entry, field, value)
#     # entry.stamp(user=user, message=message)

#     assert getattr(entry, field) == value
#     assert entry.last_modified > previous_last_modified
#     assert entry.last_modified_by != previous_last_modified_by
#     assert entry.last_modified_by == user
#     assert entry.op == entities.EntryOp.UPDATED
#     assert entry.message == message
#     assert entry.domain_events[-1] == events.EntryUpdated(
#         entity_id=entry.id,
#         entry_id=entry.entry_id,
#         repo_id=entry.repo_id,
#         user=entry.last_modified_by,
#         timestamp=entry.last_modified,
#         message=entry.message,
#         version=2,
#         body=entry.body,
#     )


# def test_resource_create_entry_from_raw():
#     resource = create_resource(
#         {
#             "resource_id": "places",
#             "resource_name": "Platser i Sverige",
#             "fields": {
#                 "name": {"type": "string", "required": True},
#                 "municipality": {
#                     "collection": True,
#                     "type": "number",
#                     "required": True,
#                 },
#                 "population": {"type": "number"},
#                 "area": {"type": "number"},
#                 "density": {"type": "number"},
#                 "code": {"type": "number", "required": True},
#             },
#             "sort": "name",
#             "id": "code",
#         }
#     )
#
#     entry = entry_lifecycle.create_entry_from_dict(resource, common_data.PLACES[0])
#
#     assert isinstance(entry, Entry)
#     assert entry.entry_id == "1"



"""Time utilities."""
import math
import struct
from datetime import datetime, timezone


def utc_now() -> float:
    """A UTC timestamp in seconds.

    This function may return the same timestamp twice.
    """
    return datetime.now(timezone.utc).timestamp()


_previous: float = utc_now()


def monotonic_utc_now() -> float:
    """A UTC timestamp in seconds.

    This function will never return the same result twice.
    """
    global _previous
    result = utc_now()
    if result <= _previous:
        result = _next_up(result)
    _previous = result
    return result


def _next_up(x: float) -> float:
    if math.isnan(x) or (math.isinf(x) and x > 0):
        return x

    if x == 0.0:
        x = 0.0

    n = struct.unpack("<q", struct.pack("<d", x))[0]
    if n >= 0:
        n += 1
    else:
        n -= 1
    return struct.unpack("<d", struct.pack("<q", n))[0]



import injector
import pytest

from karp.foundation.commands import CommandBus
from karp.foundation.events import EventBus
from karp.lex import Lex
from karp.lex_infrastructure import GenericLexInfrastructure
from karp.main.modules import CommandBusMod, EventBusMod
from karp.search import Search
from karp.search_infrastructure import SearchInfrastructure, GenericSearchInfrastructure
from karp.tests.unit.lex.adapters import InMemoryLexInfrastructure

from . import adapters


@pytest.fixture()
def search_unit_ctx() -> adapters.SearchUnitTestContext:
    container = injector.Injector(
        [
            CommandBusMod(),
            EventBusMod(),
            Search(),
            SearchInfrastructure(),
            Lex(),
            GenericLexInfrastructure(),
            InMemoryLexInfrastructure(),
            GenericSearchInfrastructure(),
            adapters.InMemorySearchInfrastructure(),
        ],
        auto_bind=False,
    )
    return adapters.SearchUnitTestContext(
        container=container,
        command_bus=container.get(CommandBus),  # type: ignore
        event_bus=container.get(EventBus),  # type: ignore
    )



from pathlib import Path

from alembic.config import Config as AlembicConfig
import alembic
import pydantic

from karp.foundation import commands


class RunMigrationsUp(pydantic.BaseModel):
    to_revision: str = "head"


class RunMigrationsDown(pydantic.BaseModel):
    to_revision: str = "base"


class RunningMigrationsBase:
    def __init__(self):
        self.alembic_cfg = AlembicConfig(Path(__file__).parent / "alembic.ini")
        self.alembic_cfg.set_main_option("script_location", str(Path(__file__).parent))


class RunningMigrationsUp(
    RunningMigrationsBase,
    commands.CommandHandler[RunMigrationsUp],
):
    def execute(self, command: RunMigrationsUp):
        alembic.command.upgrade(
            self.alembic_cfg,
            command.to_revision,
        )


class RunningMigrationsDown(
    RunningMigrationsBase, commands.CommandHandler[RunMigrationsDown]
):
    def execute(self, command: RunMigrationsDown):
        alembic.command.downgrade(
            self.alembic_cfg,
            command.to_revision,
        )



"""SQL Resource Repository"""
import logging
import typing
from typing import Dict, List, Optional, Tuple, Union
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy import sql
from sqlalchemy.orm import sessionmaker, Session

from karp.foundation.events import EventBus
from karp.lex.domain import errors, entities
from karp.lex.application import repositories
from karp.lex.domain.errors import IntegrityError, RepositoryStatusError
from karp.lex.domain.entities.resource import Resource, ResourceOp

from karp.db_infrastructure import db
from karp.db_infrastructure.sql_unit_of_work import SqlUnitOfWork
from karp.lex_infrastructure.sql.sql_models import ResourceModel
from karp.db_infrastructure.sql_repository import SqlRepository

logger = logging.getLogger(__name__)


class SqlResourceRepository(SqlRepository, repositories.ResourceRepository):
    def __init__(self, session: db.Session):
        repositories.ResourceRepository.__init__(self)
        SqlRepository.__init__(self, session=session)

    def check_status(self):
        self._check_has_session()
        try:
            self._session.execute("SELECT 1")
        except db.SQLAlchemyError as err:
            logger.exception(str(err))
            raise errors.RepositoryStatusError("Database error") from err

    @classmethod
    def primary_key(cls):
        return "resource_id"

    def _save(self, resource: Resource):
        self._check_has_session()
        resource_dto = ResourceModel.from_entity(resource)
        self._session.add(resource_dto)

    def resource_ids(self) -> List[str]:
        self._check_has_session()
        subq = (
            sql.select(
                ResourceModel.entity_id,
                sa.func.max(ResourceModel.last_modified).label("maxdate"),
            )
            .group_by(ResourceModel.entity_id)
            .subquery("t2")
        )

        stmt = sql.select(ResourceModel.resource_id).join(
            subq,
            sa.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
                ResourceModel.discarded == False,
            ),
        )
        return self._session.execute(stmt).scalars().all()
        # query = self._session.query(ResourceModel)
        # return [row.resource_id for row in query.group_by(ResourceModel.resource_id).all()]

    def _by_id(
        self, id: Union[UUID, str], *, version: Optional[int] = None
    ) -> typing.Optional[entities.Resource]:
        self._check_has_session()
        query = self._session.query(ResourceModel).filter_by(entity_id=id)
        if version:
            query = query.filter_by(version=version)
        else:
            query = query.order_by(ResourceModel.version.desc())
        resource_dto = query.first()
        return resource_dto.to_entity() if resource_dto else None

    def _by_resource_id(
        self,
        resource_id: str,
    ) -> Optional[Resource]:
        self._check_has_session()
        subq = (
            sql.select(
                ResourceModel.entity_id,
                sa.func.max(ResourceModel.last_modified).label("maxdate"),
            )
            .group_by(ResourceModel.entity_id)
            .subquery("t2")
        )

        stmt = sql.select(ResourceModel).join(
            subq,
            sa.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
                ResourceModel.resource_id == resource_id,
            ),
        )
        stmt = stmt.order_by(ResourceModel.last_modified.desc())

        query = self._session.execute(stmt).scalars()
        resource_dto = query.first()
        return resource_dto.to_entity() if resource_dto else None

    def get_active_resource(self, resource_id: str) -> Optional[Resource]:
        self._check_has_session()
        query = self._session.query(ResourceModel)
        resource_dto = query.filter_by(
            resource_id=resource_id, is_published=True
        ).one_or_none()
        return resource_dto.to_entity()

    def get_latest_version(self, resource_id: str) -> int:
        self._check_has_session()
        row = (
            self._session.query(ResourceModel)
            .order_by(ResourceModel.version.desc())
            .filter_by(resource_id=resource_id)
            .first()
        )
        if row is None:
            return 0
        return row.version

    def history_by_resource_id(
        self, resource_id: str
    ) -> typing.List[entities.Resource]:
        self._check_has_session()
        query = self._session.query(ResourceModel)
        return [
            resource_dto.to_entity()
            for resource_dto in query.filter_by(resource_id=resource_id)
            .order_by(ResourceModel.version.desc())
            .all()
        ]

    def _get_published_resources(self) -> typing.List[entities.Resource]:
        self._check_has_session()
        subq = (
            self._session.query(
                ResourceModel.resource_id,
                db.func.max(ResourceModel.last_modified).label("maxdate"),
            )
            .group_by(ResourceModel.resource_id)
            .subquery("t2")
        )
        query = self._session.query(ResourceModel).join(
            subq,
            db.and_(
                ResourceModel.resource_id == subq.c.resource_id,
                ResourceModel.last_modified == subq.c.maxdate,
                ResourceModel.is_published == True,
            ),
        )

        return [
            resource_dto.to_entity()
            for resource_dto in query
            if resource_dto is not None
        ]

    def _get_all_resources(self) -> typing.List[entities.Resource]:
        self._check_has_session()
        subq = (
            self._session.query(
                ResourceModel.resource_id,
                db.func.max(ResourceModel.last_modified).label("maxdate"),
            )
            .group_by(ResourceModel.resource_id)
            .subquery("t2")
        )
        query = self._session.query(ResourceModel).join(
            subq,
            db.and_(
                ResourceModel.resource_id == subq.c.resource_id,
                ResourceModel.last_modified == subq.c.maxdate,
            ),
        )

        return [
            resource_dto.to_entity()
            for resource_dto in query
            if resource_dto is not None
        ]

    def num_entities(self) -> int:
        self._check_has_session()
        subq = (
            self._session.query(
                ResourceModel.entity_id,
                sa.func.max(ResourceModel.last_modified).label("maxdate"),
            )
            .group_by(ResourceModel.entity_id)
            .subquery("t2")
        )
        query = self._session.query(ResourceModel).join(
            subq,
            db.and_(
                ResourceModel.entity_id == subq.c.entity_id,
                ResourceModel.last_modified == subq.c.maxdate,
                ResourceModel.discarded == False,
            ),
        )

        return query.count()

    def _resource_to_dict(self, resource: Resource) -> typing.Dict:
        return {
            "history_id": None,
            "id": resource.id,
            "resource_id": resource.resource_id,
            "version": resource.version,
            "name": resource.name,
            "config": resource.config,
            "is_published": resource.is_published,
            "last_modified": resource.last_modified,
            "last_modified_by": resource.last_modified_by,
            "message": resource.message,
            "op": resource.op,
            "discarded": resource.discarded,
        }


class SqlResourceUnitOfWork(SqlUnitOfWork, repositories.ResourceUnitOfWork):
    def __init__(
        self,
        event_bus: EventBus,
        *,
        session_factory: Optional[sessionmaker] = None,
        session: Optional[Session] = None,
    ):
        if not session and not session_factory:
            raise ValueError("Both session and session_factory cannot be None")
        SqlUnitOfWork.__init__(self, session=session)
        repositories.ResourceUnitOfWork.__init__(self, event_bus)
        self.session_factory = session_factory
        self._resources = None

    def _begin(self):
        if self._session_is_created_here:
            self._session = self.session_factory()  # type: ignore
        logger.info("using session", extra={"session": self._session})
        self._resources = SqlResourceRepository(self._session)
        return self

    @property
    def repo(self) -> SqlResourceRepository:
        if self._resources is None:
            raise RuntimeError("No resources")
        return self._resources

    # def _resource_to_row(
    #     self, resource: Resource
    # ) -> Tuple[
    #     None,
    #     UUID,
    #     str,
    #     int,
    #     str,
    #     Dict,
    #     Optional[bool],
    #     float,
    #     str,
    #     str,
    #     ResourceOp,
    #     bool,
    # ]:
    #     # config = resource.config
    #     resource.config["entry_repository_type"] = resource.entry_repository_type
    #     resource.config[
    #         "entry_repository_settings"
    #     ] = resource.entry_repository_settings
    #     return (
    #         None,
    #         resource.id,
    #         resource.resource_id,
    #         resource.version,
    #         resource.name,
    #         resource.config,
    #         resource.is_published,
    #         resource.last_modified,
    #         resource.last_modified_by,
    #         resource.message,
    #         resource.op,
    #         resource.discarded,
    #     )

    # def _row_to_resource(self, row) -> Resource:
    #     entry_repository_type = row.config.pop("entry_repository_type")
    #     entry_respsitory_settings = row.config.pop("entry_repository_settings")
    #     return Resource(
    #         entity_id=row.id,
    #         resource_id=row.resource_id,
    #         version=row.version,
    #         name=row.name,
    #         config=row.config,
    #         message=row.message,
    #         op=row.op,
    #         is_published=row.is_published,
    #         last_modified=row.last_modified,
    #         last_modified_by=row.last_modified_by,
    #         discarded=row.discarded,
    #         entry_repository_type=entry_repository_type,
    #         entry_repository_settings=entry_respsitory_settings,
    #     )



"""Constraints."""
from karp.foundation import constraints
from karp.lex.domain import errors


def valid_resource_id(name: str):
    try:
        name = constraints.length_ge("resource_id", name, 2)
        return constraints.no_space_in_str(name)
    except constraints.ConstraintsError as err:
        raise errors.InvalidResourceId(str(err)) from err



"""init db

Revision ID: 5bb29472d33f
Revises:
Create Date: 2018-12-12 13:48:19.255822

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy_utils import UUIDType
from sqlalchemy_json import NestedMutableJson

# revision identifiers, used by Alembic.
revision = "5bb29472d33f"
down_revision = None
branch_labels = None
depends_on = None


def create_resources():
    op.create_table(
        "resources",
        sa.Column("history_id", sa.Integer(), nullable=False),
        sa.Column("entity_id", UUIDType, nullable=False),
        sa.Column("resource_id", sa.String(length=32), nullable=False),
        sa.Column("resource_type", sa.String(length=32), nullable=False),
        sa.Column("entry_repo_id", UUIDType),
        sa.Column("version", sa.Integer(), nullable=False),
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.Column("config", NestedMutableJson, nullable=False),
        sa.Column("is_published", sa.Boolean, index=True, nullable=True, default=None),
        sa.Column("last_modified", sa.Float(precision=53), nullable=False),
        sa.Column("last_modified_by", sa.String(100), nullable=False),
        sa.Column("message", sa.String(100), nullable=False),
        sa.Column("op", sa.Enum("ADDED", "UPDATED", "DELETED"), nullable=False),
        sa.Column("discarded", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("history_id"),
        sa.UniqueConstraint(
            "entity_id", "version", name="entity_id_version_unique_constraint"
        ),
    )


def upgrade():
    create_resources()
    # op.create_table(
    #     "dummy_entry",
    #     sa.Column("id", sa.Integer(), nullable=False),
    #     sa.Column("body", sa.Text(), nullable=False),
    #     sa.Column("deleted", sa.Boolean(), nullable=True),
    #     sa.Column("entry_id", sa.String(length=30), nullable=False),
    #     sa.PrimaryKeyConstraint("id"),
    #     sa.UniqueConstraint("entry_id", name="entry_id_unique_constraint"),
    # )
    # op.create_table(
    #     "dummy_history",
    #     sa.Column("id", sa.Integer(), nullable=False),
    #     sa.Column("entry_id", sa.Integer(), nullable=False),
    #     sa.Column("user_id", sa.Text(), nullable=False),
    #     sa.Column(
    #         "timestamp", sa.DateTime(), server_default=sa.text("now()"), nullable=False
    #     ),
    #     sa.Column("body", sa.Text(), nullable=True),
    #     sa.Column("op", sa.Enum("ADD", "DELETE", "UPDATE"), nullable=False),
    #     sa.Column("version", sa.Integer(), nullable=False),
    #     sa.Column("message", sa.Text(), nullable=True),
    #     sa.PrimaryKeyConstraint("id"),
    #     sa.UniqueConstraint(
    #         "entry_id", "version", name="entry_id_version_unique_constraint"
    #     ),
    # )


def downgrade():
    op.drop_table("resources")
    # op.drop_table("dummy_history")
    # op.drop_table("dummy_entry")



import typing

import pydantic

from karp.foundation.value_objects import unique_id
from karp.utility import time

from karp.lex.domain import errors
from .base import Command


class CreateResource(Command):
    entity_id: unique_id.UniqueId = pydantic.Field(
        default_factory=unique_id.make_unique_id
    )
    resource_id: str
    name: str
    config: typing.Dict
    message: str
    user: str
    entry_repo_id: unique_id.UniqueId

    @classmethod
    def from_dict(
        cls,
        data: typing.Dict,
        entry_repo_id: unique_id.UniqueIdPrimitive,
        user: typing.Optional[str] = None,
        message: typing.Optional[str] = None,
    ):
        try:
            resource_id = data.pop("resource_id")
        except KeyError as exc:
            raise errors.ConstraintsError("'resource_id' is missing") from exc
        try:
            name = data.pop("resource_name")
        except KeyError as exc:
            raise errors.ConstraintsError("'resource_name' is missing") from exc

        entry_repo_id = unique_id.parse(entry_repo_id)

        return cls(
            resource_id=resource_id,
            name=name,
            config=data,
            entry_repo_id=entry_repo_id,
            user=user or "Unknown user",
            message=message or f"Resource '{resource_id}' created.",
        )


class UpdateResource(Command):
    resource_id: str
    version: int
    name: str
    config: typing.Dict
    message: str
    user: str


class PublishResource(Command):
    resource_id: str
    message: str
    user: str
    version: int


class DeleteResource(Command):
    resource_id: str
    message: str
    user: str


class SetEntryRepoId(Command):
    resource_id: str
    entry_repo_id: unique_id.UniqueId
    user: str



from .sql_entry_uows import SqlEntryUowRepository, SqlEntryUowRepositoryUnitOfWork
from .sql_entries import SqlEntryUowV1Creator, SqlEntryUowV2Creator
from .sql_resources import SqlResourceRepository, SqlResourceUnitOfWork



import pytest

from fastapi import status


class TestQueryApiInput:
    @pytest.mark.parametrize(
        "query_string",
        [
            "sort=name",
            "sort=name|asc",
            "sort=name|desc",
            "from=10",
            "size=10",
            "q=",
            "q=&sort=name&from=0&size=25",
        ],
    )
    @pytest.mark.parametrize("path", ["query", "query/split"])
    def test_valid_query_strings(self, fa_client, path: str, query_string: str):
        response = fa_client.get(f"/{path}/places?{query_string}")

        print(f"{response.json()=}")
        assert response.status_code != status.HTTP_422_UNPROCESSABLE_ENTITY



import pytest
from paradigmextract import morphparser

from karp.lex.domain.entities.entry import Entry

# from karp.lex.domain.entities.morphological_entry import (MorphologicalEntry,
#                                                     create_morphological_entry)

pytestmark = pytest.mark.skip()


def test_morph_entry_has_entry_id():
    morph_entry = create_morphological_entry(
        "pn_test", pos="pn", form_msds=[], var_insts=[], resource_id="m"
    )

    assert isinstance(morph_entry, Entry)
    assert isinstance(morph_entry, MorphologicalEntry)
    assert morph_entry.entry_id == "pn_test"


def test_morph_entry_has_function_get_inflection_table():
    morph_entry = create_morphological_entry(
        "pn_test", pos="pn", form_msds=[], var_insts=[], resource_id="m"
    )

    assert getattr(morph_entry, "get_inflection_table")


def test_morph_entry_inflect_av_1_blå():
    morph_entry = create_morphological_entry(
        "av_1_blå",
        pos="av",
        form_msds=[
            ("1", "pos indef sg u nom"),
            ("1+s", "pos indef sg u gen"),
            ("1+tt", "pos indef sg n nom"),
            ("1+tts", "pos indef sg n gen"),
            ("1+a", "pos indef pl nom"),
            ("1", "pos indef pl nom"),
            ("1+as", "pos indef pl gen"),
            ("1+s", "pos indef pl gen"),
            ("1+a", "pos def sg no_masc nom"),
            ("1", "pos def sg no_masc nom"),
            ("1+as", "pos def sg no_masc gen"),
            ("1+s", "pos def sg no_masc gen"),
            ("1+e", "pos def sg masc nom"),
            ("1+es", "pos def sg masc gen"),
            ("1+a", "pos def pl nom"),
            ("1", "pos def pl nom"),
            ("1+as", "pos def pl gen"),
            ("1+s", "pos def pl gen"),
            ("1+are", "komp nom"),
            ("1+ares", "komp gen"),
            ("1+ast", "super indef nom"),
            ("1+asts", "super indef gen"),
            ("1+aste", "super def no_masc nom"),
            ("1+astes", "super def no_masc gen"),
            ("1+aste", "super def masc nom"),
            ("1+astes", "super def masc gen"),
        ],
        var_insts=[[("1", "blå")]],
        resource_id="m",
    )

    inflection_table = morph_entry.get_inflection_table("grå")

    assert inflection_table == [
        ("pos indef sg u nom", "grå"),
        ("pos indef sg u gen", "grås"),
        ("pos indef sg n nom", "grått"),
        ("pos indef sg n gen", "gråtts"),
        ("pos indef pl nom", "gråa"),
        ("pos indef pl nom", "grå"),
        ("pos indef pl gen", "gråas"),
        ("pos indef pl gen", "grås"),
        ("pos def sg no_masc nom", "gråa"),
        ("pos def sg no_masc nom", "grå"),
        ("pos def sg no_masc gen", "gråas"),
        ("pos def sg no_masc gen", "grås"),
        ("pos def sg masc nom", "gråe"),
        ("pos def sg masc gen", "gråes"),
        ("pos def pl nom", "gråa"),
        ("pos def pl nom", "grå"),
        ("pos def pl gen", "gråas"),
        ("pos def pl gen", "grås"),
        ("komp nom", "gråare"),
        ("komp gen", "gråares"),
        ("super indef nom", "gråast"),
        ("super indef gen", "gråasts"),
        ("super def no_masc nom", "gråaste"),
        ("super def no_masc gen", "gråastes"),
        ("super def masc nom", "gråaste"),
        ("super def masc gen", "gråastes"),
    ]


def test_morph_entry_inflect_av_1_höger():
    morph_entry = create_morphological_entry(
        "av_1_höger", pos="av", form_msds=[], var_insts=[], resource_id="m"
    )

    inflection_table = morph_entry.get_inflection_table("höger")

    pass


def test_morph_entry_inflect_nn_0n_ansvar():
    morph_entry = create_morphological_entry(
        "nn_0n_ansvar",
        pos="nn",
        form_msds=[
            ("1", "sg indef nom"),
            ("1+s", "sg indef gen"),
            ("1+et", "sg def nom"),
            ("1+ets", "sg def gen"),
        ],
        var_insts=[[("1", "ansvar")]],
        resource_id="m",
    )

    inflection_table = morph_entry.get_inflection_table("apa")

    assert inflection_table == [
        ("sg indef nom", "apa"),
        ("sg indef gen", "apas"),
        ("sg def nom", "apaet"),
        ("sg def gen", "apaets"),
    ]


def test_morph_entry_inflect_nn_3u_son():
    morph_entry = create_morphological_entry(
        "nn_3u_son",
        pos="nn",
        form_msds=[
            ("1+o+2", "sg indef nom"),
            ("1+o+2+s", "sg indef gen"),
            ("1+o+2+en", "sg def nom"),
            ("1+o+2+ens", "sg def gen"),
            ("1+ö+2+er", "pl indef nom"),
            ("1+ö+2+ers", "pl indef gen"),
            ("1+ö+2+erna", "pl def nom"),
            ("1+ö+2+ernas", "pl def gen"),
        ],
        var_insts=[[("1", "s"), ("2", "n")]],
        resource_id="m",
    )

    inflection_table = morph_entry.get_inflection_table("styrelseledamot")

    assert inflection_table == [
        ("sg indef nom", "styrelseledamot"),
        ("sg indef gen", "styrelseledamots"),
        ("sg def nom", "styrelseledamoten"),
        ("sg def gen", "styrelseledamotens"),
        ("pl indef nom", "styrelseledamöter"),
        ("pl indef gen", "styrelseledamöters"),
        ("pl def nom", "styrelseledamöterna"),
        ("pl def gen", "styrelseledamöternas"),
    ]


@pytest.mark.xfail(reason="can't match")
def test_morph_entry_inflect_nn_2u_bövel():
    morph_entry = create_morphological_entry(
        "nn_2u_bövel",
        pos="nn",
        form_msds=[("1+e+2", "sg indef nom")],
        var_insts=[[("1", "böv"), ("2", "l")]],
        resource_id="m",
    )
    inflection_table = morph_entry.get_inflection_table("sommar")

    assert inflection_table == [("sg indef nom", "sommar")]


def test_morph_entry_inflect_nn_0n_syre():
    morph_entry = create_morphological_entry(
        "nn_0n_syre",
        pos="nn",
        form_msds=[
            ("1", "sg indef nom"),
            ("1+s", "sg indef gen"),
            ("1+t", "sg def nom"),
            ("1+ts", "sg def gen"),
        ],
        var_insts=[[("1", "syre")]],
        resource_id="m",
    )

    inflection_table = morph_entry.get_inflection_table("vete")

    assert inflection_table == [
        ("sg indef nom", "vete"),
        ("sg indef gen", "vetes"),
        ("sg def nom", "vetet"),
        ("sg def gen", "vetets"),
    ]


@pytest.mark.skip(reason="How shall it be used.")
def test_morph_entry_inflect_ab():
    morph_entry = create_morphological_entry(
        "ab_1_illa", pos="ab", form_msds=[("")], var_insts=[], resource_id="m"
    )

    inflection_table = morph_entry.get_inflection_table("")


def test_morph_entry_inflect_vb_2a_känna():
    morph_entry = create_morphological_entry(
        "vb_2a_känna",
        pos="vb",
        form_msds=[
            ("1+ner", "pres ind aktiv"),
            ("1+nes", "pres ind s-form"),
            ("1+ns", "pres ind s-form"),
            ("1+de", "pret ind aktiv"),
            ("1+des", "pret ind s-form"),
            ("1+n", "imper"),
            ("1+na", "inf aktiv"),
            ("1+nas", "inf s-form"),
            ("1+t", "sup aktiv"),
            ("1+ts", "sup s-form"),
        ],
        var_insts=[[("1", "kän")]],
        resource_id="m",
    )

    inflection_table = morph_entry.get_inflection_table("bränna")

    assert inflection_table == [
        ("pres ind aktiv", "bränner"),
        ("pres ind s-form", "brännes"),
        ("pres ind s-form", "bränns"),
        ("pret ind aktiv", "brände"),
        ("pret ind s-form", "brändes"),
        ("imper", "bränn"),
        ("inf aktiv", "bränna"),
        ("inf s-form", "brännas"),
        ("sup aktiv", "bränt"),
        ("sup s-form", "bränts"),
    ]


def test_morph_entry_inflect_vb_1a_laga():
    morph_entry = create_morphological_entry(
        "vb_1a_laga",
        pos="vb",
        form_msds=[
            ("1+r", "pres ind aktiv"),
            ("1+s", "pres ind s-form"),
            ("1+de", "pret ind aktiv"),
            ("1+des", "pret ind s-form"),
            ("1", "imper"),
            ("1", "inf aktiv"),
            ("1+s", "inf s-form"),
            ("1+t", "sup aktiv"),
            ("1+ts", "sup s-form"),
        ],
        var_insts=[[("1", "laga")]],
        resource_id="m",
    )

    inflection_table = morph_entry.get_inflection_table("jobba")

    assert inflection_table == [
        ("pres ind aktiv", "jobbar"),
        ("pres ind s-form", "jobbas"),
        ("pret ind aktiv", "jobbade"),
        ("pret ind s-form", "jobbades"),
        ("imper", "jobba"),
        ("inf aktiv", "jobba"),
        ("inf s-form", "jobbas"),
        ("sup aktiv", "jobbat"),
        ("sup s-form", "jobbats"),
    ]


def test_morph_entry_vb_1s_andas():
    morph_entry = create_morphological_entry(
        "vb_1s_andas",
        pos="vb",
        form_msds=[
            ("1+2", "pres ind s-form"),
            ("1+de+2", "pret ind s-form"),
            ("1+2", "imper"),
            ("1+2", "inf s-form"),
            ("1+t+2", "sup s-form"),
        ],
        var_insts=[[("1", "anda"), ("2", "s")]],
        resource_id="m"
        # var_insts=[[("1", "anda"), ("2", "s")]],
    )

    assert morph_entry.tags == ("inf aktiv", "inf s-form")

    inflection_table = morph_entry.get_inflection_table("fightas")

    variables = morphparser.eval_baseform(
        morph_entry.paradigm, "fightas", morph_entry.tags
    )

    assert variables == ["fighta", "s"]
    assert inflection_table == [
        ("pres ind s-form", "fightas"),
        ("pret ind s-form", "fightades"),
        ("imper", "fightas"),
        ("inf s-form", "fightas"),
        ("sup s-form", "fightats"),
    ]



import pytest

from typer import Typer
from typer.testing import CliRunner
from karp import lex

from karp.main import AppContext


class TestCliResourceLifetime:
    def test_help(self, runner: CliRunner, cliapp: Typer):
        result = runner.invoke(cliapp, ["resource", "--help"])
        assert result.exit_code == 0

    def test_create_and_publish_repo(
        self,
        runner: CliRunner,
        cliapp: Typer,
        app_context: AppContext,
    ):
        result = runner.invoke(
            cliapp, ["entry-repo", "create", "karp/tests/data/config/lexlex.json"]
        )
        print(f"{result.stdout=}")
        assert result.exit_code == 0

        entry_repo_repo = app_context.container.get(lex.ReadOnlyEntryRepoRepositry)  # type: ignore [misc]
        entry_repo = entry_repo_repo.get_by_name("lexlex")
        assert entry_repo is not None

        result = runner.invoke(
            cliapp,
            [
                "resource",
                "create",
                "karp/tests/data/config/lexlex.json",
                "--entry-repo-id",
                str(entry_repo.entity_id),
            ],
        )
        print(f"{result.stdout=}")
        assert result.exit_code == 0

        resource_repo = app_context.container.get(lex.ReadOnlyResourceRepository)  # type: ignore [misc]
        assert resource_repo.get_by_resource_id("lexlex") is not None

    @pytest.mark.xfail(reason="not ready")
    def test_update_entry_repo(
        self,
        runner: CliRunner,
        cliapp: Typer,
        app_context: AppContext,
    ):
        result = runner.invoke(
            cliapp, ["entry-repo", "create", "karp/tests/data/config/lexlex.json"]
        )
        print(f"{result.stdout=}")
        assert result.exit_code == 0

        entry_repo_repo = app_context.container.get(lex.ReadOnlyEntryRepoRepositry)  # type: ignore [misc]
        entry_repo = entry_repo_repo.get_by_name("lexlex")
        assert entry_repo is not None

        result = runner.invoke(
            cliapp,
            [
                "resource",
                "set-entry-repo",
                "lexlex",
                str(entry_repo.entity_id),
            ],
        )
        print(f"{result.stdout=}")
        assert result.exit_code == 0

        resource_repo = app_context.container.get(lex.ReadOnlyResourceRepository)  # type: ignore [misc]
        resource_lexlex = resource_repo.get_by_resource_id("lexlex")
        assert resource_lexlex.entry_repository_id == entry_repo.entity_id



import json
from typing import Callable, Dict, List, Optional, Tuple

import pytest  # pyre-ignore
from fastapi import status

from karp import auth
from karp.lex.application.queries import EntryViews
from karp.tests.common_data import MUNICIPALITIES, PLACES
from karp.tests.utils import add_entries, get_json


def extract_names(entries):
    return [entry["entry"]["name"] for entry in entries["hits"]]


def extract_names_set(entries):
    return {entry["entry"]["name"] for entry in entries["hits"]}


def _test_path(
    client,
    path: str,
    expected_result: List[str],
    *,
    access_token: Optional[auth.AccessToken] = None,
    headers: Optional[Dict] = None,
) -> None:
    if access_token:
        if headers is None:
            headers = access_token.as_header()
        else:
            headers.extend(access_token.as_header())
    kwargs = {"headers": headers} if headers else {}

    entries = get_json(client, path, **kwargs)

    names = extract_names(entries)
    print("names = {names}".format(names=names))

    assert len(names) == len(expected_result)

    for name in names:
        assert name in expected_result

    for expected in expected_result:
        assert expected in names


def _test_path_has_expected_length(
    client,
    path: str,
    expected_length: int,
    *,
    access_token: Optional[auth.AccessToken] = None,
    headers: Optional[Dict] = None,
) -> None:
    if access_token:
        if headers is None:
            headers = access_token.as_header()
        else:
            headers.extend(access_token.as_header())
    kwargs = {"headers": headers} if headers else {}

    entries = get_json(client, path, **kwargs)

    assert len(entries["hits"]) == expected_length


def _test_against_entries(
    client,
    path: str,
    field: str,
    predicate: Callable,
    expected_n_hits: int = None,
    *,
    access_token: Optional[auth.AccessToken] = None,
    headers: Optional[Dict] = None,
) -> None:
    if access_token:
        if headers is None:
            headers = access_token.as_header()
        else:
            headers.extend(access_token.as_header())
    kwargs = {"headers": headers} if headers else {}

    entries = get_json(client, path, **kwargs)
    names = extract_names(entries)

    if field.endswith(".raw"):
        field = field[:-4]

    if expected_n_hits:
        assert len(entries["hits"]) == expected_n_hits
    else:
        assert len(names) == sum(
            bool(field in entry and predicate(entry[field])) for entry in PLACES
        )

    num_names_to_match = len(names)
    for entry in PLACES:
        if field in entry and predicate(entry[field]):
            assert entry["name"] in names
            num_names_to_match -= 1
    assert num_names_to_match == 0


def _test_against_entries_general(
    client,
    path: str,
    fields: Tuple[str, str],
    predicate: Callable,
    expected_n_hits: int = None,
    *,
    access_token: Optional[auth.AccessToken] = None,
    headers: Optional[Dict] = None,
) -> None:
    if access_token:
        if headers is None:
            headers = access_token.as_header()
        else:
            headers.update(access_token.as_header())
    kwargs = {"headers": headers} if headers else {}

    entries = get_json(client, path, **kwargs)
    names = extract_names(entries)

    print("names = {names}".format(names=names))
    for i, field in enumerate(fields):
        if field.endswith(".raw"):
            fields[i] = field[:-4]

    for field in fields:
        print("field = {field}".format(field=field))

    num_names_to_match = len(names)
    for entry in PLACES:
        print("entry = {entry}".format(entry=entry))
        if predicate(entry, fields):
            print("entry '{entry}' satisfied predicate".format(entry=entry))
            assert entry["name"] in names
            num_names_to_match -= 1
    assert num_names_to_match == 0

    if expected_n_hits:
        assert len(entries["hits"]) == expected_n_hits


class TestQuery:
    def test_route_exist(self, fa_data_client):
        response = fa_data_client.get("/query/places")
        print(f"{response.json()=}")
        assert response.status_code != status.HTTP_404_NOT_FOUND


def test_query_no_q(
    fa_data_client,
    read_token: auth.AccessToken,
    app_context,
):
    resource = "places"
    entries = get_json(
        fa_data_client,
        f"/query/{resource}",
        headers=read_token.as_header(),
    )

    names = extract_names(entries)

    entry_views = app_context.container.get(EntryViews)  # type: ignore [misc]
    expected_result = {}
    expected_total = entry_views.get_total(resource)
    print(f"entries = {entries}")
    assert entries["total"] == expected_total
    assert len(names) == expected_total
    print("names = {}".format(names))

    # for entry in PLACES:
    #     assert entry["name"] in names

    return

    for i, name in enumerate(names):
        print("name = {}".format(name))
        print("  entry = {}".format(entries["hits"][i]["entry"]))
        if name in ["Grund test", "Hambo", "Alhamn"]:
            print("Testing 'larger_place' for '{}' ...".format(name))
            print("  entry = {}".format(entries["hits"][i]["entry"]))
            assert "larger_place" in entries["hits"][i]["entry"]
            assert "v_larger_place" in entries["hits"][i]["entry"]
        if name == "Alvik":
            print("Testing 'smaller_places' for '{}' ...".format(name))
            print("  entry = {}".format(entries["hits"][i]["entry"]))
            assert "v_smaller_places" in entries["hits"][i]["entry"]
            assert (
                entries["hits"][i]["entry"]["v_smaller_places"][0]["name"] == "Bjurvik"
            )
        elif name in ["Botten test", "Alhamn"]:
            print("Testing 'smaller_places' for '{}' ...".format(name))
            print("  entry = {}".format(entries["hits"][i]["entry"]))
            assert "v_smaller_places" in entries["hits"][i]["entry"]


def test_query_split(
    fa_data_client,
    read_token: auth.AccessToken,
):
    resources = ["places", "municipalities"]
    entries = get_json(
        fa_data_client,
        "/query/split/{}".format(",".join(resources)),
        headers=read_token.as_header(),
    )

    entry_views = fa_data_client.app.state.app_context.container.get(EntryViews)
    expected_result = {}
    for resource in resources:
        expected_result[resource] = entry_views.get_total(resource)
    assert entries["distribution"] == expected_result
    # assert entries["distribution"] == {"municipalities": 3, "places": 22}


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "queries,expected_result",
    [
        (["equals|population|4133", "equals|area|50000"], ["Botten test"]),
        (["regexp|name|.*bo.*", "equals|area|50000"], ["Hambo", "Botten test"]),
        (["regexp|name|.*bo.*", "equals|area|50000", "missing|density"], ["Hambo"]),
    ],
)
def test_and(
    fa_data_client,
    queries: List[str],
    expected_result: List[str],
):
    query = "/query/places?q=and({queries})".format(queries="||".join(queries))
    _test_path(fa_data_client, query, expected_result)


@pytest.mark.parametrize(
    "field,value",
    [
        ("name", "grund"),
        ("name", "Grund"),
        ("name.raw", "Grund"),
        # pytest.param(
        #     "population",
        #     3122,
        #     ["Grund test"],
        #     marks=pytest.mark.skip(
        #         reason="Can't search with regex on LONG fields"),
        # ),
        # ("|and|name|v_larger_place.name|", "vi", ["Bjurvik"]),
        # pytest.param(
        #     "|and|name|v_smaller_places.name|",
        #     "and|Al|vi",
        #     ["Alvik"],
        #     marks=pytest.mark.xfail(reason="regex can't handle complex"),
        # ),
        # pytest.param(
        #     "|not|name|",
        #     "test",
        #     ["Hambo", "Alhamn", "Bjurvik", "Bjurvik2"],
        #     marks=pytest.mark.xfail(reason="?"),
        # ),
        # ("|or|name|v_smaller_places.name|",
        #  "Al", ["Alvik", "Alhamn", "Hambo"]),
        # ("name", "|and|un|es", ["Grund test"]),
        # pytest.param(
        #     "name",
        #     "|or|vi|bo",
        #     ["Alvik", "Rutvik", "Bjurvik", "Botten test", "Hambo", "Bjurvik2"],
        #     marks=pytest.mark.xfail(reason="?"),
        # ),
    ],
)
def test_contains(
    fa_data_client,
    field: str,
    value,
    app_context,
):
    query = f"/query/places?q=contains|{field}|{value}"
    entry_views = app_context.container.get(EntryViews)  # type: ignore [misc]
    expected_result = []
    real_field = field.split(".raw")[0]
    if real_field == field:
        analyzed_value = value
    else:
        analyzed_value = value.lower()
    print(f"{value=} {analyzed_value=}")
    for entry in entry_views.all_entries("places"):
        print(f"{entry=}")
        print(f"{entry.entry=}")
        print(f"{entry.entry.get(real_field, None)=}")
        if analyzed_value in entry.entry.get(real_field, "").lower():
            expected_result.append(entry.entry["name"])
    print(f"{expected_result=}")
    _test_path(fa_data_client, query, expected_result)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "fields,values,expected_result",
    [(("name", "v_larger_place.name"), ("vi", "vi"), ["Bjurvik"])],
)
def test_contains_and_separate_calls(
    fa_data_client,
    fields: Tuple,
    values: Tuple,
    expected_result: List[str],
):
    names = set()
    for field, value in zip(fields, values):
        query = f"/query/places?q=contains|{field}|{value}"
        entries = get_json(
            fa_data_client,
            query,
        )
        if not names:
            print("names is empty")
            names = extract_names_set(entries)
        else:
            names = names & extract_names_set(entries)

    assert len(names) == len(expected_result)
    for name in names:
        assert name in expected_result
    for expected in expected_result:
        assert expected in names


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,value,expected_result",
    [
        ("name", "est", ["Grund test", "Botten test"]),
        ("name", "unds", ["Grunds"]),
        ("name", "grund", ["Grund test"]),
        # pytest.param("population", 3122, [
        #              "Grund test"], marks=pytest.mark.skip),
        # ("|and|name|v_smaller_places.name|", "vik", ["Alvik"]),
        # ("|and|name|v_larger_place.name|", "vik", ["Bjurvik"]),
        # pytest.param(
        #     "|and|name|v_smaller_places.name|",
        #     "and|Al|vi",
        #     ["Alvik"],
        #     marks=pytest.mark.skip(reason="regex can't handle complex"),
        # ),
        # pytest.param(
        #     "|not|name|",
        #     "vik",
        #     [
        #         "Botten test",  # through larger_place
        #         "Alvik",  # through smaller_places
        #         "Bjurvik",  # through larger_place
        #     ],
        #     marks=pytest.mark.xfail(reason="Too restrictive"),
        # ),
        # ("|or|name|v_smaller_places.name|",
        #  "otten", ["Botten test", "Bjurvik"]),
        # ("name", "|and|und|est", ["Grund test"]),
        # ("name", "|or|vik|bo", ["Alvik", "Rutvik", "Bjurvik", "Hambo"]),
    ],
)
def test_endswith(fa_data_client, field: str, value, expected_result: List[str]):
    query = f"/query/places?q=endswith|{field}|{value}"
    _test_path(fa_data_client, query, expected_result)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,value,expected_result",
    [
        ("name", "grunds", ["Grunds"]),
        ("name", "Grunds", ["Grunds"]),
        ("name", "Grund test", ["Grund test"]),
        ("name.raw", "Grund test", ["Grund test"]),
        # ("name", "|and|grund|test", ["Grund test"]),
        ("density", 7, ["Botten test"]),
        # ("|and|population|area|", 6312, ["Alvik"]),
        # pytest.param(
        #     "|and|area||or|population|density|",
        #     6312,
        #     ["Alvik", "Grund test"],
        #     marks=pytest.mark.xfail(
        #         reason="Current query dsl can't handle this."),
        # ),
        # pytest.param(
        #     "|not|area|",
        #     6312,
        #     ["Alvik", "Grunds"],
        #     marks=pytest.mark.xfail(reason="Too restrictive."),
        # ),
        # (
        #     "|or|population|area|",
        #     6312,
        #     ["Alhamn", "Alvik", "Bjurvik", "Grund test", "Grunds"],
        # ),
        # ("name", "|and|botten|test", ["Botten test"]),
        # ("population", "|or|6312|3122", ["Alvik", "Grund test", "Grunds"]),
    ],
)
def test_equals(fa_data_client, field: str, value, expected_result: List[str]):
    query = f"/query/places?q=equals|{field}|{value}"
    _test_path(fa_data_client, query, expected_result)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,expected_result",
    [
        # (
        #     "|and|v_larger_place|v_smaller_places",
        #     ["Bjurvik", "Hambo", "Botten test", "Alhamn", "Grund test"],
        # ),
        ("name", [])
    ],
)
def test_exists(fa_data_client, field: str, expected_result: List[str]):
    query = f"/query/places?q=exists|{field}"
    _test_path(fa_data_client, query, expected_result)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,expected_result",
    [
        pytest.param(
            "|and|.*test|Gr.*",
            [
                "Grund test",
                "Alhamn",  # through smaller_places
                "Bjurvik2",  # through larger_place
            ],
            marks=pytest.mark.xfail(reason="?"),
        ),
        # ("|not|Gr.*", ["Botten test", "Hambo", "Alvik", "Rutvik", "Bjurvik"]),
        pytest.param(
            "|or|.*test|Gr.*",
            [
                "Grund test",
                "Grunds",
                "Botten test",
                "Hambo",  # through larger_place
                "Alhamn",  # through smaller_places
                "Bjurvik",  # through smaller_places
                "Bjurvik2",  # through larger_place
            ],
            marks=pytest.mark.xfail(reason="?"),
        ),
        pytest.param(
            "Grunds?",
            ["Grund test", "Grunds", "Bjurvik2", "Alhamn"],
            marks=pytest.mark.xfail(reason="?"),
        ),
    ],
)
def test_freergxp(fa_data_client, field: str, expected_result: List[str]):
    query = f"/query/places?q=freergxp|{field}"
    _test_path(fa_data_client, query, expected_result)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,expected_result",
    [
        pytest.param(
            "grund",
            [
                "Grund test",
                "Grunds",
                "Alhamn",  # through smaller_places
                "Bjurvik2",  # through larger_place
            ],
            marks=pytest.mark.xfail(reason="?"),
        ),
        ("3122", ["Grund test"]),
        (3122, ["Grund test"]),
        # (
        #     "|and|botten|test",
        #     [
        #         "Botten test",
        #         "Hambo",  # through larger_place
        #         "Bjurvik",  # through smaller_places
        #     ],
        # ),
        pytest.param(
            "|or|botten|test",
            [
                "Botten test",
                "Hambo",  # through smaller_places
                "Grund test",
                "Alhamn",  # through smaller_places
                "Bjurvik",  # through smaller_places
                "Bjurvik2",  # through larger_places
            ],
            marks=pytest.mark.xfail(reason="?"),
        ),
    ],
)
def test_freetext(fa_data_client, field: str, expected_result: List[str]):
    query = f"/query/places?q=freetext|{field}"
    _test_path(fa_data_client, query, expected_result)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,value,expected_n_hits",
    [
        ("population", (4132,), 4),
        ("area", (20000,), 11),
        ("name", ("alvik", "Alvik"), 16),
        pytest.param(
            "name",
            ("Alvik",),
            2,
            marks=pytest.mark.skip(
                reason="'name' is lower case, so greater than Uppercase returns all."
            ),
        ),
        pytest.param(
            "name",
            ("B",),
            2,
            marks=pytest.mark.skip(
                reason="'name' is lower case, so greater than Uppercase returns all."
            ),
        ),
        ("name.raw", ("Alvik",), 20),
        ("name.raw", ("B",), 20),
        pytest.param(
            "name",
            ("r", "R"),
            2,
            marks=pytest.mark.xfail(
                reason="'name' is tokenized, so greater than matches second word."
            ),
        ),
        pytest.param(
            "name",
            ("R",),
            2,
            marks=pytest.mark.xfail(
                reason="'name' is lower case, so greater than Uppercase returns all."
            ),
        ),
        ("name.raw", ("r",), 5),
        ("name.raw", ("R",), 15),
    ],
)
def test_gt(fa_data_client, field, value, expected_n_hits):
    query = f"/query/places?q=gt|{field}|{value[0]}"
    _test_path_has_expected_length(fa_data_client, query, expected_n_hits)
    # _test_against_entries(fa_data_client, query, field, lambda x: value[-1] < x)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,value,expected_n_hits",
    [
        ("population", (4132,), 5),
        ("area", (20000,), 12),
        ("name", ("alvik", "Alvik"), 17),
        pytest.param(
            "name",
            ("Alvik",),
            2,
            marks=pytest.mark.skip(
                reason="'name' is lower case, so greater than Uppercase returns all."
            ),
        ),
        pytest.param(
            "name",
            ("B",),
            2,
            marks=pytest.mark.skip(
                reason="'name' is lower case, so greater than Uppercase returns all."
            ),
        ),
        ("name.raw", ("Alvik",), 21),
        ("name.raw", ("B",), 20),
    ],
)
def test_gte(fa_data_client, field, value, expected_n_hits):
    query = f"/query/places?q=gte|{field}|{value[0]}"
    _test_path_has_expected_length(fa_data_client, query, expected_n_hits)
    # _test_against_entries(fa_data_client, query, field, lambda x: value[-1] <= x)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,value,expected_n_hits",
    [
        ("population", (4132,), 10),
        ("area", (20000,), 4),
        ("name", ("alvik", "Alvik"), 5),
        pytest.param(
            "name",
            ("Alvik",),
            2,
            marks=pytest.mark.skip(
                reason="'name' is lower case, so lesser than Uppercase returns all."
            ),
        ),
        pytest.param(
            "name",
            ("B",),
            2,
            marks=pytest.mark.skip(
                reason="'name' is lower case, so lesser than Uppercase returns all."
            ),
        ),
        ("name.raw", ("Alvik",), 1),
        ("name.raw", ("B",), 2),
    ],
)
def test_lt(fa_data_client, field, value, expected_n_hits: int):
    query = f"/query/places?q=lt|{field}|{value[0]}"
    _test_path_has_expected_length(fa_data_client, query, expected_n_hits)
    # _test_against_entries(fa_data_client, query, field, lambda x: x < value[-1])


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,value,expected_n_hits",
    [
        ("population", (4132,), 11),
        ("area", (20000,), 5),
        ("name", ("alvik", "Alvik"), 6),
        pytest.param(
            "name",
            ("Alvik",),
            2,
            marks=pytest.mark.skip(
                reason="'name' is lower case, so lesser than Uppercase returns all."
            ),
        ),
        pytest.param(
            "name",
            ("B",),
            2,
            marks=pytest.mark.skip(
                reason="'name' is lower case, so lesser than Uppercase returns all."
            ),
        ),
        ("name.raw", ("Alvik",), 2),
        ("name.raw", ("B",), 2),
    ],
)
def test_lte(fa_data_client, field, value, expected_n_hits: int):
    query = f"/query/places?q=lte|{field}|{value[0]}"
    _test_path_has_expected_length(fa_data_client, query, expected_n_hits)
    # _test_against_entries(fa_data_client, query, field, lambda x: x <= value[-1])


@pytest.mark.skip()
@pytest.mark.parametrize(
    "op,fields,value,expected_n_hits",
    [
        ("gt", ("population", "area"), (6212,), 2),
        (
            "gt",
            ("name", "v_smaller_places.name"),
            ("bjurvik",),
            4,
        ),
        ("gte", ("population", "area"), (6212,), 3),
        (
            "gte",
            ("name", "v_smaller_places.name"),
            ("bjurvik",),
            5,
        ),
        ("lt", ("population", "area"), (6313,), 4),
        ("lt", ("name", "v_smaller_places.name"), ("c",), 2),
        ("lte", ("population", "area"), (6312,), 4),
        ("lte", ("name", "v_smaller_places.name"), ("bjurvik",), 1),
    ],
)
def test_binary_range_1st_arg_and(
    fa_data_client,
    op: str,
    fields: Tuple[str, str],
    value: Tuple,
    expected_n_hits: int,
):
    query = f"/query/places?q={op}||and|{fields[0]}|{fields[1]}||{value[0]}"
    _test_path_has_expected_length(fa_data_client, query, expected_n_hits)
    # if expected_result:
    #     _test_path(fa_data_client, query, expected_result)

    # elif op == "gt":
    #     _test_against_entries_general(
    #         fa_data_client,
    #         query,
    #         fields,
    #         lambda entry, fields: all(
    #             f in entry and value[-1] < entry[f] for f in fields
    #         ),
    #     )
    # elif op == "gte":
    #     _test_against_entries_general(
    #         fa_data_client,
    #         query,
    #         fields,
    #         lambda entry, fields: all(
    #             f in entry and value[-1] <= entry[f] for f in fields
    #         ),
    #     )
    # elif op == "lt":
    #     _test_against_entries_general(
    #         fa_data_client,
    #         query,
    #         fields,
    #         lambda entry, fields: all(
    #             f in entry and entry[f] < value[-1] for f in fields
    #         ),
    #     )
    # elif op == "lte":
    #     _test_against_entries_general(
    #         fa_data_client,
    #         query,
    #         fields,
    #         lambda entry, fields: all(
    #             f in entry and entry[f] <= value[-1] for f in fields
    #         ),
    #     )
    # else:
    #     pytest.fail(msg="Unknown range operator '{op}'".format(op=op))


# @pytest.mark.parametrize(
#     "op,fields,value,expected_n_hits",
#     [
#         ("gt", ("population", "area"), (6212,), 16),
#         (
#             "gt",
#             ("name", "v_smaller_places.name"),
#             ("bjurvik",),
#             16,
#         ),
#         ("gte", ("population", "area"), (6212,), 16),
#         (
#             "gte",
#             ("name", "v_smaller_places.name"),
#             ("bjurvik",),
#             17,
#         ),
#         ("lt", ("population", "area"), (6212,), 12),
#         (
#             "lt",
#             ("name", "v_smaller_places.name"),
#             ("bjurvik",),
#             8,
#         ),
#         ("lte", ("population", "area"), (6212,), 13),
#         (
#             "lte",
#             ("name", "v_smaller_places.name"),
#             ("bjurvik",),
#             9,
#         ),
#     ],
# )
# def test_binary_range_1st_arg_or(
#     fa_data_client,
#     op: str,
#     fields: Tuple,
#     value: Tuple,
#     expected_n_hits: int,
# ):
#     query = f"/query/places?q={op}||or|{fields[0]}|{fields[1]}||{value[0]}"
#     _test_path_has_expected_length(fa_data_client, query, expected_n_hits)

# elif op == "gt":
#     _test_against_entries_general(
#         fa_data_client,
#         query,
#         fields,
#         lambda entry, fields: any(
#             f not in entry or value[-1] < entry[f] for f in fields
#         ),
#     )
# elif op == "gte":
#     _test_against_entries_general(
#         fa_data_client,
#         query,
#         fields,
#         lambda entry, fields: any(
#             f not in entry or value[-1] <= entry[f] for f in fields
#         ),
#     )
# elif op == "lt":
#     _test_against_entries_general(
#         fa_data_client,
#         query,
#         fields,
#         lambda entry, fields: any(
#             f in entry and entry[f] < value[-1] for f in fields
#         ),
#     )
# elif op == "lte":
#     _test_against_entries_general(
#         fa_data_client,
#         query,
#         fields,
#         lambda entry, fields: any(
#             f in entry and entry[f] <= value[-1] for f in fields
#         ),
#     )
# else:
#     pytest.fail(msg=f"Unknown range operator '{op}'")


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,lower,upper,expected_n_hits",
    [
        ("population", (3812,), (4133,), 1),
        pytest.param("area", (6312,), (50000,), 1, marks=pytest.mark.xfail(reason="?")),
        # ("name", ("alhamn", "Alhamn"), ("bjurvik", "Bjurvik"), 2),
        pytest.param("name", ("b", "B"), ("h", "H"), 1, marks=pytest.mark.xfail),
        pytest.param("name", ("Alhamn",), ("Bjurvik",), 1, marks=pytest.mark.xfail),
        pytest.param("name", ("B",), ("H",), 1, marks=pytest.mark.xfail),
        ("name.raw", ("Alhamn",), ("Bjurvik",), 1),
        pytest.param(
            "name.raw", ("B",), ("H",), 4, marks=pytest.mark.xfail(reason="?")
        ),
    ],
)
def test_and_gt_lt(fa_data_client, field, lower, upper, expected_n_hits):
    query = f"/query/places?q=and(gt|{field}|{lower[0]}||lt|{field}|{upper[0]})"
    print(f"testing query='{query}'")
    _test_against_entries(
        fa_data_client,
        query,
        field,
        lambda x: lower[-1] < x < upper[-1],
        expected_n_hits,
    )


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "query,expected_n_hits", [("and(gt|name|alhamn||lt|name|bjurvik)", 2)]
)
def test_and_gt_lt_expected_length(fa_data_client, query: str, expected_n_hits: int):
    path = f"/query/places?q={query}"
    _test_path_has_expected_length(fa_data_client, path, expected_n_hits)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,expected_length",
    [
        ("density", 8),
        # ("|and|density|population", 7),
        # (
        #     "|not|density",
        #     14,
        # ),
        # ("|or|density|population", 8),
        # (
        #     "|or|density|population|v_smaller_places",
        #     17,
        # ),
    ],
)
def test_missing(fa_data_client, field: str, expected_length: int):
    query = f"/query/places?q=missing|{field}"
    _test_path_has_expected_length(fa_data_client, query, expected_length)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "queries,expected_length",
    [
        (
            "freetext|botten",
            19,
        ),
        ("freergxp|.*test", 17),
        ("freergxp|.*test||freergxp|.*vik", 14),
    ],
)
def test_not(fa_data_client, queries: str, expected_length: int):
    query = f"/query/places?q=not({queries})"
    _test_path_has_expected_length(fa_data_client, query, expected_length)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "queries,expected_result",
    [
        (["equals|population|3122", "equals|population|4132"], ["Hambo", "Grund test"]),
        (
            ["equals|population|3122", "equals|population|4132", "endswith|name|est"],
            ["Hambo", "Grund test", "Botten test"],
        ),
    ],
)
def test_or(fa_data_client, queries: List[str], expected_result: List[str]):
    query = f"/query/places?q=or({'||'.join(queries)})"
    _test_path(fa_data_client, query, expected_result)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,value,expected_result",
    [
        ("name", "grun.*", ["Grund test", "Grunds"]),
        ("name", "Grun.*", ["Grund test", "Grunds"]),
        ("name", "Grunds?", ["Grunds"]),
        ("name", "grunds?", ["Grund test", "Grunds"]),
        ("name", "Grun.*est", ["Grund test"]),
        # ("name", "|and|grun.*|.*est", ["Grund test"]),
        # ("name", "|or|grun.*|.*est", ["Grund test", "Grunds", "Botten test"]),
        # ("|and|name|v_larger_place.name|", ".*vik", ["Bjurvik"]),
        # ("|or|name|v_larger_place.name|", "al.*n", ["Grund test", "Alhamn"]),
        # ("|not|name|", "Al.*", ["Grund test", "Hambo", "Bjurvik"]),
    ],
)
def test_regexp(fa_data_client, field: str, value, expected_result: List[str]):
    query = f"/query/places?q=regexp|{field}|{value}"
    _test_path(fa_data_client, query, expected_result)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "field,value,expected_result",
    [
        ("name", "grun", ["Grund test", "Grunds"]),
        ("name", "Grun", ["Grund test", "Grunds"]),
        ("name", "tes", ["Grund test", "Botten test"]),
        # ("name", "|and|grun|te", ["Grund test"]),
        # ("name", "|or|grun|te", ["Grund test", "Grunds", "Botten test"]),
        # ("|and|name|v_larger_place.name|", "b", ["Botten test"]),
        # ("|and|name|v_larger_place.name|", "B", ["Botten test"]),
        # ("|or|name|v_larger_place.name|", "alh", ["Grund test", "Alhamn"]),
        # ("|not|name|", "Al", ["Grund test", "Hambo", "Bjurvik"]),
    ],
)
def test_startswith(fa_data_client, field: str, value, expected_result: List[str]):
    query = f"/query/places?q=startswith|{field}|{value}"
    _test_path(fa_data_client, query, expected_result)


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize(
    "query_str,expected_length",
    [
        ("contains|name|2", 6),
        pytest.param(
            "contains|name||not|test",
            20,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
        pytest.param(
            "contains|name||not|test|bo",
            19,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
        ("endswith|name|2", 1),
        pytest.param(
            "endswith|name||not|vik",
            19,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
        pytest.param(
            "equals|area||not|6312",
            18,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
        ("exists|density", 14),
        pytest.param(
            "exists||and|density|population",
            14,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
        pytest.param(
            "exists||or|density|population",
            15,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
        pytest.param(
            "exists||not|density",
            8,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
        pytest.param(
            "freergxp||not|Gr.*",
            19,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
        pytest.param(
            "freetext||not|botten",
            19,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
        pytest.param(
            "freetext||not|botten|test",
            17,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
        pytest.param(
            "regexp|name||not|.*est",
            20,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
        pytest.param(
            "startswith|name||not|te",
            20,
            marks=pytest.mark.skip(reason="currently not supported"),
        ),
    ],
)
def test_response_has_correct_length(
    fa_data_client, query_str: str, expected_length: int
):
    query = f"/query/places?q={query_str}"
    print(f"testing query='{query}'")
    _test_path_has_expected_length(fa_data_client, query, expected_length)


# @pytest.mark.xfail(reason="no protected stuff")
# def test_protected(client_with_data_scope_module):
#     response = client_with_data_scope_module.get("/municipalities/query")
#     names = response
#     assert "403 FORBIDDEN" in names


@pytest.mark.xfail(reason="unstable")
def test_pagination_explicit_0_5(fa_data_client):
    # client = init_data(client_with_data_f, es, 30)
    resource = "places"
    json_data = get_json(
        fa_data_client,
        f"/query/{resource}?from=0&size=5",
    )
    assert "hits" in json_data
    assert "distribution" in json_data
    print(f"json_data = {json_data}")
    assert len(json_data["hits"]) == 5

    hit_3 = json_data["hits"][3]

    response = fa_data_client.get(
        f"/query/{resource}?from=3&size=5&lexicon_stats=false",
    )
    assert response.status_code == 200
    json_data = response.json()
    assert json_data
    assert "hits" in json_data
    assert "distribution" not in json_data
    assert len(json_data["hits"]) == 5

    assert json_data["hits"][0]["id"] == hit_3["id"]


# #
# #
# # def test_pagination_default_size(fa_data_client):
# #     client = init_data(client_with_data_f, es, 30)
# #     resource = 'places'
# #     response = client.get('/{}/query?from=0'.format(resource))
# #     assert response.status == '200 OK'
# #     assert response.is_json
# #     json_data = response.get_json()
# #     assert len(json_data['hits']) == 25
# #
# #
# # def test_pagination_default_from(fa_data_client):
# #     client = init_data(client_with_data_f, es, 50)
# #     resource = 'places'
# #     response = client.get('/{}/query?size=45'.format(resource))
# #     assert response.status == '200 OK'
# #     assert response.is_json
# #     json_data = response.get_json()
# #     assert len(json_data['hits']) == 45


@pytest.mark.xfail(reason="unstable")
def test_pagination_fewer(fa_data_client):
    # client = init_data(client_with_data_f, es, 5)
    resource = "places"
    response = fa_data_client.get(
        f"/query/{resource}?from=10",
    )
    assert response.status_code == 200
    json_data = response.json()
    print(f"json_data = {json_data}")
    assert len(json_data["hits"]) == json_data["total"] - 10


# def test_resource_not_existing(fa_data_client):
#     response = fa_data_client.get("/asdf/query")
#     assert response.status_code == 400
#     assert (
#         'Resource is not searchable: "asdf"'
#         == json.loads(response.data.decode())["error"]
#     )


# def init_data(client, es_status_code, no_entries):
#     if es_status_code == "skip":
#         pytest.skip("elasticsearch disabled")
#     client_with_data = client(use_elasticsearch=True)

#     for i in range(0, no_entries):
#         entry = {"code": i, "name": "name", "municipality": [1]}
#         client_with_data.post(
#             "places/add",
#             data=json.dumps({"entry": entry}),
#             content_type="application/json",
#         )
#     return client_with_data


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize("endpoint", ["query", "query/split"])
@pytest.mark.parametrize(
    "query",
    [
        (""),
        ("freetext|eat my shorts"),
    ],
)
def test_distribution_in_result(fa_data_client, query: str, endpoint: str):
    result = get_json(
        fa_data_client,
        f"/{endpoint}/places?{f'q={query}&' if query else ''}lexicon_stats=true",
    )

    assert "distribution" in result


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize("endpoint", ["query"])
def test_sorting(fa_data_client, endpoint: str):
    result = get_json(
        fa_data_client,
        f"/{endpoint}/places?sort=population|desc",
    )

    assert (
        result["hits"][0]["entry"]["population"]
        >= result["hits"][1]["entry"]["population"]
    )


@pytest.mark.xfail(reason="unstable")
@pytest.mark.parametrize("fields", [(["population"])])
def test_query_include_fields(fa_data_client, fields: List[str]) -> None:
    result = get_json(
        fa_data_client,
        f"/query/places?{'&'.join((f'include_field={field}' for field in fields))}",
    )

    for entry in (entry["entry"] for entry in result["hits"]):
        for field in fields:
            assert field in fields



class Node:
    def __init__(self, type_, arity: int, value=None):
        self.type = type_
        self.arity = arity
        self.value = value
        self.children = []

    def __repr__(self):
        return "<Node {t} {v} {c}>".format(t=self.type, v=self.value, c=self.children)

    def _pprint(self, level):
        fill = " "
        print(
            "{indent} Node {t} {v}".format(
                indent=fill * level, t=self.type, v=self.value
            )
        )
        for child in self.children:
            child._pprint(level + 1)

    def pprint(self, level: int = 0):
        self._pprint(level)

    def add_child(self, child):
        self.children.append(child)

    def n_children(self) -> int:
        return len(self.children)

    def gen_stream(self):
        yield self
        for child in self.children:
            yield from child.gen_stream()


def create_unary_node(type_, value=None):
    return Node(type_, 1, value)


def create_binary_node(type_, value=None):
    return Node(type_, 2, value)



import pkg_resources

plugins = {}


def init():
    for entry_point in pkg_resources.iter_entry_points("karp.plugins"):
        plugins[entry_point.name] = entry_point.load()
        plugins[entry_point.name].init()



"""Handles all sql db connections."""
import collections
import json
from typing import Any, Dict, Optional

import attr
import sqlalchemy
from sqlalchemy import (
    JSON,
    Column,
    Enum,
    ForeignKey,
    Integer,
    MetaData,
    String,
    Table,
    Text,
    Unicode,
    and_,
    event,
    exc,
    func,
    or_,
)
from sqlalchemy.dialects.mysql import DOUBLE
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.mutable import Mutable
from sqlalchemy.orm import aliased, mapper, relationship
from sqlalchemy.orm.session import Session, sessionmaker
from sqlalchemy.schema import (
    ForeignKeyConstraint,
    PrimaryKeyConstraint,
    UniqueConstraint,
)
from sqlalchemy.sql import delete, insert, update
from sqlalchemy.types import VARCHAR, Boolean, Float, Time, TypeDecorator
from sqlalchemy_json import NestedMutableJson
from sqlalchemy_utils import UUIDType
from karp.db_infrastructure.types import ULIDType


__all__ = ["ULIDType"]

# engine = sqlalchemy.create_engine(config.DB_URL, echo=True)

metadata = MetaData()

Base = declarative_base(metadata=metadata)

# SessionLocal = sessionmaker(bind=engine)


@attr.s(auto_attribs=True)
class SQLEngineSessionfactory:
    engine: Engine = attr.ib()
    session_factory: sessionmaker = attr.ib()
    metadata: MetaData = attr.ib()

    @classmethod
    def from_db_uri(cls, db_uri: str):
        engine = sqlalchemy.create_engine(db_uri)
        session_factory = sessionmaker(bind=engine)
        metadata_ = MetaData()
        return cls(engine, session_factory, metadata_)


_db_handlers: Dict[str, SQLEngineSessionfactory] = dict()


def create_session(db_uri: str) -> Session:
    _assure_in_db_handlers(db_uri)
    return _db_handlers[db_uri].session_factory()


def get_engine(db_uri: str) -> Engine:
    _assure_in_db_handlers(db_uri)
    return _db_handlers[db_uri].engine


def _assure_in_db_handlers(db_uri: str):
    if db_uri not in _db_handlers:
        _db_handlers[db_uri] = SQLEngineSessionfactory.from_db_uri(db_uri)


def _metadata(db_uri: str) -> MetaData:
    _assure_in_db_handlers(db_uri)
    return _db_handlers[db_uri].metadata


def get_table(table_name: str) -> Optional[Table]:
    return metadata.tables.get(table_name)


# def map_class_to_some_table(cls, table: Table, entity_name: str, **mapper_kw):
#     """Use the EntityName-pattern to map a cls to several tables.

#     See https://github.com/sqlalchemy/sqlalchemy/wiki/EntityName for more information.

#     Arguments:
#         table {Table} -- the table to map to
#         entity_name {str} -- the name for the new class

#     Keyword arguments:
#     mapper_kw -- any keyword arguments passed to mapper

#     Raises:
#         ValueError: when?

#     Returns:
#         Any -- the new class constructed
#     """
#     newcls = type(entity_name, (cls,), {})
#     mapper(newcls, table, **mapper_kw)
#     return newcls


# _metadata = MetaData()


# class JsonEncodedDict(TypeDecorator):
#     """Represent an immutable dictionary as a json-encoded-string."""

#     impl = VARCHAR

#     def process_bind_param(self, value, dialect):
#         if value is not None:
#             value = json.dumps(value)
#         return value

#     def process_result_value(self, value, dialect):
#         if value is not None:
#             value = json.loads(value)
#         return value


# class MutableDict(Mutable, dict):
#     """Tracking changes for sql."""

#     @classmethod
#     def coerce(cls, key, value):
#         """Convert plain dicts to MutableDict."""
#         if not isinstance(value, MutableDict):
#             if isinstance(value, dict):
#                 return cls(value)

#             # this call will raise ValueError
#             return Mutable.coerce(key, value)
#         else:
#             return value

#     def __setitem__(self, key, value):
#         """Detect dictionary set events and emit changed event."""
#         dict.__setitem__(self, key, value)
#         self.changed()

#     def __delitem__(self, key, value):
#         """Detect dictionary del events and emit changed event."""
#         dict.__delitem__(self, key, value)
#         self.changed()


# MutableDict.associate_with(JsonEncodedDict)


# class FalseEncodedAsNullBoolean(TypeDecorator):
#     impl = Boolean

#     def process_bind_param(self, value, dialect):
#         if value is False:
#             value = None
#         return value

#     def process_result_value(self, value, dialect):
#         if value is None:
#             return False
#         return value


# class MutableBool(Mutable):
#     @classmethod
#     def coerce(cls, key, value):
#         if isinstance(value, MutableBool):
#             return value
#
#         if isinstance(value, bool):
#             return cls(value)
#
#         return Mutable.coerce(key, value)
#
#     def __setattr__(self, name, value):
#         object.__setattr__(self, name, value)
#         self.changed()


# MutableBool.associate_with(FalseEncodedAsNullBoolean)



import logging
import random
from re import I
from typing import Optional
import urllib.parse
from urllib.parse import urlunparse

import urllib3
import urllib3.exceptions

from starlette.requests import Request
from starlette.responses import Response
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.types import ASGIApp


logger = logging.getLogger(__name__)

http = urllib3.PoolManager()


def get_remote_address(req: Request) -> str:
    return req.scope["client"][0]


class MatomoMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        idsite: str,
        matomo_url: str,
        access_token: Optional[str] = None,
        *,
        assume_https: bool = True,
    ) -> None:
        super().__init__(app)
        self.idsite = idsite
        self.matomo_url = matomo_url
        self.assume_https = assume_https
        self._access_token = access_token

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        headers = {}
        for k, v in request.headers.items():
            headers[k.lower()] = v
        scope = request.scope
        if "x-forwarded-server" in headers:
            server = headers["x-forwarded-server"]
        else:
            host, port = scope["server"]
            server = f"{host}:{port}"
        root_path = scope.get("root_path")
        path = scope["path"]
        if root_path:
            path = f"{root_path}{path}"
        url = urlunparse(
            (
                "https" if self.assume_https else str(scope["scheme"]),
                server,
                path,
                "",
                str(scope["query_string"]),
                "",
            )
        )
        logger.info(
            "got request",
            extra={
                "client": request.scope["client"],
                "server": request.scope["server"],
                "url": url,
                "headers": headers,
            },
        )
        cip = get_remote_address(request)
        response = await call_next(request)

        params_that_require_token = {}
        if self._access_token:
            params_that_require_token = {"token_auth": self._access_token, "cip": cip}
        else:
            logger.warning("Not recording client ip", extra={"cip": cip})

        tracking_dict = {
            "idsite": self.idsite,
            "url": url,
            "rec": 1,
            "rand": random.getrandbits(32),
            "apiv": 1,
            "ua": headers.get("user-agent"),
            "lang": headers.get("accept-lang"),
            **params_that_require_token,
        }
        tracking_params = urllib.parse.urlencode(tracking_dict)
        tracking_url = f"{self.matomo_url}?{tracking_params}"
        try:
            logger.debug("Making tracking call", extra={"url": tracking_url})
            r = http.request("GET", tracking_url)
            logger.debug(
                "tracking call", extra={"status_code": r.status, "data": r.data}
            )
        except urllib3.exceptions.HTTPError:
            logger.exception("Error tracking view")
        return response



import pytest

# from karp.search.domain import errors, index
# from karp.services import entry_query
# from karp.tests import random_refs

# from .adapters import bootstrap_test_app

pytestmark = pytest.mark.skip()


class TestSearchIds:
    def test_cannot_search_non_existent_resource(self):
        bus = bootstrap_test_app()
        with pytest.raises(errors.ResourceNotFound):
            entry_query.search_ids("non_existing", "entry", bus.ctx)

    def test_cannot_search_non_published_resource(self):
        bus = bootstrap_test_app()
        bus.handle(random_refs.make_create_resource_command("existing"))
        with pytest.raises(errors.ResourceNotPublished):
            entry_query.search_ids("existing", "entry", bus.ctx)


class TestQuery:
    def test_cannot_search_non_existent_resource(self):
        bus = bootstrap_test_app()
        with pytest.raises(errors.ResourceNotFound):
            query_request = index.QueryRequest(resource_ids="non_existing")
            entry_query.query(query_request, bus.ctx)

    def test_cannot_search_non_published_resource(self):
        bus = bootstrap_test_app()
        bus.handle(random_refs.make_create_resource_command("existing"))
        with pytest.raises(errors.ResourceNotPublished):
            query_request = index.QueryRequest(resource_ids="existing")
            entry_query.query(query_request, bus.ctx)


class TestQuerySplit:
    def test_cannot_search_non_existent_resource(self):
        bus = bootstrap_test_app()
        with pytest.raises(errors.ResourceNotFound):
            query_request = index.QueryRequest(resource_ids="non_existing")
            entry_query.query_split(query_request, bus.ctx)

    def test_cannot_search_non_published_resource(self):
        bus = bootstrap_test_app()
        bus.handle(random_refs.make_create_resource_command("existing"))
        with pytest.raises(errors.ResourceNotPublished):
            query_request = index.QueryRequest(resource_ids="existing")
            entry_query.query_split(query_request, bus.ctx)



import copy
import uuid
from unittest import mock

import pytest

from karp.lex.domain import errors, events
from karp.lex.domain.errors import (
    ConsistencyError,
    ConstraintsError,
    DiscardedEntityError,
)
from karp.lex.domain.entities.resource import (
    Release,
    Resource,
    ResourceOp,
    create_resource,
)
from karp.foundation.value_objects import unique_id

from . import factories
from .factories import random_resource


def test_create_resource_creates_resource():
    resource_id = "test_resource"
    name = "Test resource"
    conf = {
        "resource_id": resource_id,
        "resource_name": name,
        "sort": ["baseform"],
        "fields": {"baseform": {"type": "string", "required": True}},
    }
    with mock.patch("karp.utility.time.utc_now", return_value=12345):
        resource = create_resource(
            conf,
            created_by="kristoff@example.com",
            entry_repo_id=unique_id.make_unique_id(),
        )

    assert isinstance(resource, Resource)
    assert isinstance(resource.entity_id, unique_id.UniqueIdType)
    assert resource.entity_id == unique_id.parse(str(resource.entity_id))
    assert resource.version == 1
    assert resource.resource_id == resource_id
    assert resource.name == name
    assert not resource.discarded
    assert not resource.is_published
    assert "resource_id" not in resource.config
    assert "resource_name" not in resource.config
    assert "sort" in resource.config
    assert resource.config["sort"] == conf["sort"]
    assert "fields" in resource.config
    assert resource.config["fields"] == conf["fields"]
    assert int(resource.last_modified) == 12345
    assert resource.message == "Resource added."
    assert resource.op == ResourceOp.ADDED
    assert resource.domain_events[-1] == events.ResourceCreated(
        timestamp=resource.last_modified,
        entity_id=resource.id,
        entry_repo_id=resource.entry_repository_id,
        resource_id=resource.resource_id,
        name=resource.name,
        config=resource.config,
        user=resource.last_modified_by,
        message=resource.message,
    )


def test_resource_update_changes_last_modified_and_version():
    resource_id = "test_resource"
    name = "Test resource"
    conf = {
        "resource_id": resource_id,
        "resource_name": name,
        "sort": ["baseform"],
        "fields": {"baseform": {"type": "string", "required": True}},
        "entry_repository_type": None,
        "entry_repository_settings": {},
    }

    resource = factories.ResourceFactory(
        name=copy.deepcopy(name), config=copy.deepcopy(conf)
    )
    previous_last_modified = resource.last_modified
    previous_version = resource.version

    conf["fields"]["new"] = {"type": "string"}
    resource.update(name="new name", config=conf, user="Test", version=1)

    assert resource.last_modified > previous_last_modified
    assert resource.last_modified_by == "Test"
    assert resource.version == (previous_version + 1)
    assert resource.domain_events[-1] == events.ResourceUpdated(
        timestamp=resource.last_modified,
        entity_id=resource.id,
        entry_repo_id=resource.entry_repository_id,
        resource_id=resource.resource_id,
        name=resource.name,
        config=resource.config,
        user=resource.last_modified_by,
        message=resource.message,
        version=resource.version,
    )


# def test_resource_add_new_release_creates_release():
#     resource = factories.ResourceFactory()

#     previous_last_modified = resource.last_modified

#     resource.add_new_release(name="v1.0.0", user="Admin", description="")

#     assert len(resource.releases) == 1
#     assert resource.releases[0].name == "v1.0.0"
#     assert resource.releases[0].publication_date == resource.last_modified
#     assert resource.releases[0].description == ""
#     assert resource.releases[0].root.id == resource.id
#     assert resource.last_modified > previous_last_modified
#     assert resource.last_modified_by == "Admin"
#     assert resource.message == "Release 'v1.0.0' created."
#     assert resource.version == 2


# def test_resource_release_with_name_on_discarded_raises_discarded_entity_error():
#     resource = factories.ResourceFactory()

#     resource.discard(user="Test", message="Discard")

#     assert resource.discarded

#     with pytest.raises(DiscardedEntityError):
#         resource.release_with_name("test")


# def test_resource_add_new_release_on_discarded_raises_discarded_entity_error():
#     resource = factories.ResourceFactory()

#     resource.discard(user="Test", message="Discard")

#     assert resource.discarded

#     with pytest.raises(DiscardedEntityError):
#         resource.add_new_release(name="test", user="TEST", description="")


# @pytest.mark.skip()
# def test_resource_add_new_release_with_invalid_name_raises_constraints_error():
#     resource = factories.ResourceFactory()

#     with pytest.raises(ConstraintsError):
#         resource.add_new_release(name="", user="Test", description="")


# @pytest.mark.skip()
# def test_resource_new_release_added_with_wrong_version_raises_consistency_error():
#     resource = factories.ResourceFactory()
#     event = Resource.NewReleaseAdded(
#         entity_id=resource.id,
#         entity_version=12,
#     )
#     with pytest.raises(ConsistencyError):
#         event.mutate(resource)


# @pytest.mark.skip()
# def test_resource_new_release_added_with_wrong_last_modified_raises_consistency_error():
#     resource = factories.ResourceFactory()
#     event = Resource.NewReleaseAdded(
#         entity_id=resource.id, entity_version=resource.version, entity_last_modified=2
#     )
#     with pytest.raises(ConsistencyError):
#         event.mutate(resource)


# def test_release_created_has_id():
#     release = Release(
#         entity_id="e", name="e-name", publication_date=12345.0, description="ee"
#     )

#     assert release.id == "e"
#     assert release.name == "e-name"
#     assert release.publication_date == 12345.0
#     assert release.description == "ee"
#     assert release.root.id == release.id


# def test_release_created_w_resource_has_id():
#     resource = factories.ResourceFactory()
#     release = Release(
#         entity_id="e",
#         name="e-name",
#         publication_date=12345.0,
#         description="ee",
#         aggregate_root=resource,
#     )

#     assert release.id == "e"
#     assert release.name == "e-name"
#     assert release.publication_date == 12345.0
#     assert release.description == "ee"
#     assert release.root.id == resource.id


def test_resource_has_entry_json_schema():
    resource = factories.ResourceFactory()

    json_schema = resource.entry_json_schema

    assert json_schema["type"] == "object"
    assert "baseform" in json_schema["properties"]


@pytest.mark.parametrize(
    "field,value",
    [
        # ("resource_id", "new..1"),
        # ("config", {"b": "r"}),
        ("name", "New name"),
    ],
)
def test_discarded_resource_has_event(field, value):
    resource = random_resource()
    resource.discard(user="alice@example.org", message="bad", timestamp=123.45)
    assert resource.discarded
    assert resource.domain_events[-1] == events.ResourceDiscarded(
        entity_id=resource.id,
        # entry_repo_id=resource.entry_repository_id,
        resource_id=resource.resource_id,
        user=resource.last_modified_by,
        timestamp=resource.last_modified,
        message=resource.message,
        version=2,
        name=resource.name,
        config=resource.config,
    )
    with pytest.raises(errors.DiscardedEntityError):
        setattr(resource, field, value)


def test_published_resource_has_event():
    resource = random_resource()
    previous_version = resource.version
    resource.publish(
        user="kristoff@example.com", message="publish", version=resource.version
    )
    assert resource.is_published
    assert resource.version == (previous_version + 1)
    assert resource.domain_events[-1] == events.ResourcePublished(
        entity_id=resource.id,
        entry_repo_id=resource.entry_repository_id,
        resource_id=resource.resource_id,
        user=resource.last_modified_by,
        timestamp=resource.last_modified,
        version=resource.version,
        config=resource.config,
        name=resource.name,
        message=resource.message,
    )



from .entry_repos import SqlListEntryRepos, SqlReadOnlyEntryRepoRepositry
from .generic_entries import (
    GenericEntryViews,
    GenericEntryQuery,
    GenericGetEntryDiff,
    GenericGetEntryHistory,
    GenericGetHistory,
)
from .generic_network import GenericGetReferencedEntries
from .generic_resources import GenericGetEntryRepositoryId
from .resources import (
    SqlGetPublishedResources,
    SqlGetResources,
    SqlReadOnlyResourceRepository,
)



import abc
import typing

import pydantic
from sb_json_tools import jsondiff

from karp import errors as karp_errors
from karp.foundation.value_objects import unique_id
from karp.lex.application.repositories import (
    ResourceUnitOfWork,
    EntryUowRepositoryUnitOfWork,
)
from karp.lex.domain.entities.entry import EntryOp


# pylint: disable=unsubscriptable-object
class EntryDto(pydantic.BaseModel):
    # entry_id: str
    # entity_id: typing.Optional[unique_id.UniqueId] = None
    entity_id: unique_id.UniqueIdStr
    resource: str
    version: int
    entry: typing.Dict
    last_modified: float
    last_modified_by: str


class EntryDiffRequest(pydantic.BaseModel):
    resource_id: str
    # entry_id: str
    entity_id: unique_id.UniqueIdStr
    from_version: typing.Optional[int] = None
    to_version: typing.Optional[int] = None
    from_date: typing.Optional[float] = None
    to_date: typing.Optional[float] = None
    entry: typing.Optional[typing.Dict] = None


class EntryHistoryRequest(pydantic.BaseModel):
    resource_id: str
    user_id: typing.Optional[str] = None
    entry_id: typing.Optional[unique_id.UniqueIdStr] = None
    from_date: typing.Optional[float] = None
    to_date: typing.Optional[float] = None
    from_version: typing.Optional[int] = None
    to_version: typing.Optional[int] = None
    current_page: int = 0
    page_size: int = 100


class EntryDiffDto(pydantic.BaseModel):
    diff: typing.Any
    from_version: typing.Optional[int]
    to_version: typing.Optional[int]


class HistoryDto(pydantic.BaseModel):
    timestamp: float
    message: str
    # entry_id: str
    entity_id: unique_id.UniqueIdStr
    version: int
    op: EntryOp
    user_id: str
    diff: list[dict]


class GetEntryDiff(abc.ABC):
    @abc.abstractmethod
    def query(self, request: EntryDiffRequest) -> EntryDiffDto:
        pass


class GetEntryHistory(abc.ABC):
    @abc.abstractmethod
    def query(
        self,
        resource_id: str,
        entity_id: unique_id.UniqueIdStr,
        # entry_id: str,
        version: typing.Optional[int],
    ) -> EntryDto:
        pass


class GetHistoryDto(pydantic.BaseModel):
    history: list[HistoryDto]
    total: int


class GetHistory(abc.ABC):
    @abc.abstractmethod
    def query(self, request: EntryHistoryRequest) -> GetHistoryDto:
        pass


class EntryViews(abc.ABC):
    @abc.abstractmethod
    def get_by_id(
        self,
        resource_id: str,
        entity_id: unique_id.UniqueIdStr,
    ) -> EntryDto:
        pass

    @abc.abstractmethod
    def get_by_id_optional(
        self,
        resource_id: str,
        entity_id: unique_id.UniqueIdStr,
    ) -> typing.Optional[EntryDto]:
        pass

    # @abc.abstractmethod
    # def get_by_entry_id(
    #     self,
    #     resource_id: str,
    #     entry_id: str,
    # ) -> EntryDto:
    #     pass

    # @abc.abstractmethod
    # def get_by_entry_id_optional(
    #     self,
    #     resource_id: str,
    #     entry_id: str,
    # ) -> typing.Optional[EntryDto]:
    #     pass

    @abc.abstractmethod
    def get_total(self, resource_id: str) -> int:
        pass

    @abc.abstractmethod
    def get_by_referenceable(
        self, resource_id: str, filters
    ) -> typing.Iterable[EntryDto]:
        pass

    @abc.abstractmethod
    def all_entries(self, resource_id: str) -> typing.Iterable[EntryDto]:
        pass



import abc
import typing

import pydantic

from .entries import EntryDto


class ReferenceDto(pydantic.BaseModel):
    resource_id: str
    resource_version: typing.Optional[int]
    entry: EntryDto


class GetReferencedEntries(abc.ABC):
    @abc.abstractmethod
    def query(
        self,
        resource_id: str,
        entry_id: str,
    ) -> typing.Iterable[ReferenceDto]:
        pass



import typing

import pydantic

from karp.foundation.value_objects import unique_id
from karp.foundation import commands
from karp.utility import time


class Command(pydantic.BaseModel, commands.Command):
    timestamp: float = pydantic.Field(default_factory=time.utc_now)


class ReindexResource(Command):
    resource_id: str



import logging
from typing import Optional

from fastapi import Depends, HTTPException, status
from fastapi.param_functions import Header
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer, SecurityScopes
from fastapi.security import utils as security_utils
import logging

# from karp import bootstrap, services
from karp import auth, lex

# from karp.auth.auth import auth
from karp.errors import ClientErrorCodes, KarpError
from karp.auth.domain.auth_service import AuthService
from karp.auth_infrastructure import (
    JWTAuthService,
    LexGetResourcePermissions,
    LexIsResourceProtected,
)
from . import lex_deps
from .fastapi_injector import inject_from_req


auth_scheme = HTTPBearer()

logger = logging.getLogger(__name__)


def bearer_scheme(authorization=Header(None)):
    if not authorization:
        return None
    scheme, credentials = security_utils.get_authorization_scheme_param(authorization)
    if not (scheme and credentials):
        return None
    return HTTPAuthorizationCredentials(scheme=scheme, credentials=credentials)


def get_resource_permissions(
    query: lex.GetPublishedResources = Depends(lex_deps.get_published_resources),
) -> auth.GetResourcePermissions:
    return LexGetResourcePermissions(query)


# def get_resource_permissions(conn: Connection = Depends(database.get_connection)) -> GetResourcePermissions:
#     return


def get_is_resource_protected(
    repo: lex.ReadOnlyResourceRepository = Depends(lex_deps.get_resources_read_repo),
) -> auth.IsResourceProtected:
    return LexIsResourceProtected(repo)


def get_auth_service(
    config: auth.AuthServiceConfig = Depends(inject_from_req(auth.AuthServiceConfig)),
    query: auth.IsResourceProtected = Depends(get_is_resource_protected),
) -> auth.AuthService:
    return JWTAuthService(
        pubkey_path=config.pubkey_path,
        is_resource_protected=query,
    )


def get_current_user(
    security_scopes: SecurityScopes,
    credentials: Optional[HTTPAuthorizationCredentials] = Depends(bearer_scheme),
    auth_service: auth.AuthService = Depends(get_auth_service),
) -> Optional[auth.User]:
    if not credentials:
        return None
    if security_scopes.scopes:
        authenticate_value = f'Bearer scope="{security_scopes.scope_str}"'
    else:
        authenticate_value = "Bearer"
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": authenticate_value},
        # code=ClientErrorCodes.NOT_PERMITTED,
    )
    try:
        logger.debug(
            "Calling auth_service",
            extra={"credentials": credentials},
        )
        return auth_service.authenticate(credentials.scheme, credentials.credentials)
    except KarpError:
        logger.exception("error authenticating")
        raise credentials_exception


get_user_optional = get_current_user


def get_user(
    security_scopes: SecurityScopes,
    credentials: HTTPAuthorizationCredentials = Depends(auth_scheme),
    auth_service: auth.AuthService = Depends(get_auth_service),
) -> auth.User:
    if security_scopes.scopes:
        authenticate_value = f'Bearer scope="{security_scopes.scope_str}"'
    else:
        authenticate_value = "Bearer"
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": authenticate_value},
        # code=ClientErrorCodes.NOT_PERMITTED,
    )
    try:
        logger.debug(
            "Calling auth_service",
            extra={"credentials": credentials},
        )
        return auth_service.authenticate(credentials.scheme, credentials.credentials)
    except KarpError:
        raise credentials_exception



import json
from typing import Optional

from tabulate import tabulate
import typer

from karp import lex
from karp.foundation.value_objects import UniqueId
from karp.foundation.commands import CommandBus
from karp.foundation.value_objects.unique_id import UniqueIdStr
from karp.lex.application.queries import ListEntryRepos
from karp.lex.domain.commands import CreateEntryRepository
from karp.cliapp.typer_injector import inject_from_ctx


subapp = typer.Typer()


@subapp.command()
def create(infile: typer.FileBinaryRead, ctx: typer.Context):
    try:
        data = json.load(infile)
    except Exception as err:
        typer.echo(f"Error reading file '{infile.name}': {str(err)}")
        raise typer.Exit(123)
    create_entry_repo = CreateEntryRepository.from_dict(
        data,
        user="local admin",
    )

    bus = inject_from_ctx(CommandBus, ctx)

    bus.dispatch(create_entry_repo)

    typer.echo(
        "Entry repository '{name}' with id '{id}' created.".format(
            name=create_entry_repo.name,
            id=create_entry_repo.entity_id,
        )
    )


@subapp.command()
def delete(
    entity_id: str,  # TODO: use UniqueIdStr when supported,
    ctx: typer.Context,
    user: Optional[str] = typer.Option(None),
):

    # bus = inject_from_ctx(CommandBus, ctx)

    # delete_entry_repo = DeleteEntryRepo(
    #     entity_id=entity_id,
    #     user=user or "local admin"
    # )
    # typer.echo(f"Entry repository with id '{entity_id}' deleted.")
    typer.echo("not yet supported")


@subapp.command()
def list(ctx: typer.Context):
    query = inject_from_ctx(ListEntryRepos, ctx)
    typer.echo(
        tabulate(
            [
                [entry_repo.name, entry_repo.entity_id, entry_repo.repository_type]
                for entry_repo in query.query()
            ],
            headers=["name", "id", "repository_type"],
        )
    )


@subapp.command()
def show(ctx: typer.Context, name: str):
    repo = inject_from_ctx(lex.ReadOnlyEntryRepoRepositry, ctx)
    entry_repo = repo.get_by_name(name)
    if entry_repo:
        typer.echo(tabulate(((key, value) for key, value in entry_repo.dict().items())))
    else:
        typer.echo("No such entry-repo")
        raise typer.Exit(2)


def init_app(app: typer.Typer) -> None:
    app.add_typer(subapp, name="entry-repo")



import abc
from typing import Optional

import pydantic
from karp.search.application.repositories import IndexEntry
from karp.foundation.value_objects import unique_id


class PreviewEntryInputDto(pydantic.BaseModel):
    resource_id: str
    entry: dict
    user: str


class EntryPreviewDto(pydantic.BaseModel):
    entry: IndexEntry


class PreviewEntry(abc.ABC):
    @abc.abstractmethod
    def query(self, input_dto: PreviewEntryInputDto) -> EntryPreviewDto:
        ...



import typer

from karp.main.migrations import use_cases as migration_usecases


subapp = typer.Typer()


@subapp.command(name="up")
def migrations_up(
    ctx: typer.Context,
):
    uc = migration_usecases.RunningMigrationsUp()
    uc.execute(migration_usecases.RunMigrationsUp())


@subapp.command(name="down")
def migrations_down(
    ctx: typer.Context,
):
    uc = migration_usecases.RunningMigrationsDown()
    uc.execute(migration_usecases.RunMigrationsDown())


def init_app(app: typer.Typer) -> None:
    app.add_typer(subapp, name="db")



"""Handle of unique ids.

Borrowed from https://bitbucket.org/sixty-north/d5-kanban-python
"""
import uuid
import typing

import pydantic
import ulid

# UniqueId = uuid.UUID
# UniqueIdType = uuid.UUID
# typing_UniqueId = uuid.UUID
UniqueId = ulid.ULID
UniqueIdType = ulid.ULID
typing_UniqueId = ulid.ULID


def make_unique_id(timestamp: typing.Optional[float] = None) -> UniqueId:
    """Make a new UniqueId."""
    return ulid.new() if timestamp is None else ulid.from_timestamp(timestamp)


def to_unique_id(phrase: typing.Any) -> UniqueId:
    if isinstance(phrase, str):
        return ulid.from_str(phrase)
    raise ValueError()


parse = ulid.parse


class UniqueIdStr(str):
    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(examples=["01BJQMF54D093DXEAWZ6JYRPAQ"])

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, UniqueId):
            return str(v)
        elif not isinstance(v, str):
            raise TypeError("string or UniqueId required")

        if len(v) != 26:
            raise ValueError("invalid uniqueid format")

        return cls(v)

    def __repr__(self) -> str:
        return f"UniqueIdStr({super().__repr__()})"


UniqueIdPrimitive = typing.Union[ulid.api.api.ULIDPrimitive, UniqueIdStr]


# def parse(value: UniqueIdPrimitive) -> UniqueId:
#     if isinstance(value, Uni)



"""Helpers for converting.
"""
from typing import Callable, List


def str2list(delimiter: str = None) -> Callable[[str], List[str]]:
    """Return a function that splits a string using {delimiter} provided

    Keyword Arguments:
        delimiter {str} -- the delimiter to split by, defaults to ' ' (default: {None})

    Returns:
        Callable[[str], List[str]] -- the function to do the spliting.
    """

    def result(string: str) -> List[str]:
        return string.split(delimiter)

    return result



import factory


from karp.search.domain import commands


class ReindexResourceFactory(factory.Factory):
    class Meta:
        model = commands.ReindexResource

    resource_id = factory.Faker("word")



"""Module for jwt-based authentication."""
from pathlib import Path
from typing import Dict, List, Optional

import jwt
import jwt.exceptions as jwte  # pyre-ignore
import pydantic
import logging


from karp.foundation import value_objects
from karp.auth.application.queries import IsResourceProtected
from karp.auth.domain.errors import ExpiredToken, TokenError, InvalidTokenPayload
from karp.auth.domain.entities.user import User
from karp.auth import AuthService, AuthServiceConfig


logger = logging.getLogger(__name__)


class JWTMeta(pydantic.BaseModel):
    iss: str
    aud: Optional[str]
    iat: float
    exp: float


class JWTCreds(pydantic.BaseModel):
    sub: str
    levels: Dict
    scope: Optional[Dict]


class JWTPayload(JWTMeta, JWTCreds):
    pass


def load_jwt_key(path: Path) -> str:
    with open(path) as fp:
        return fp.read()


class JWTAuthServiceConfig(AuthServiceConfig):
    def __init__(self, pubkey_path: str):
        self._pubkey_path = Path(pubkey_path)

    @property
    def pubkey_path(self) -> Path:
        return self._pubkey_path


class JWTAuthService(AuthService):
    def __init__(
        self, pubkey_path: Path, is_resource_protected: IsResourceProtected
    ) -> None:
        self._jwt_key = load_jwt_key(pubkey_path)
        self.is_resource_protected = is_resource_protected
        logger.debug("JWTAuthenticator created")

    def authenticate(self, _scheme: str, credentials: str) -> User:
        logger.debug("authenticate called", extra={"credentials": credentials})

        try:
            user_token = jwt.decode(
                credentials, key=self._jwt_key, algorithms=["RS256"], leeway=5
            )
        except jwte.ExpiredSignatureError as exc:
            raise ExpiredToken() from exc
        except jwte.DecodeError as exc:
            raise TokenError() from exc

        try:
            payload = JWTPayload(**user_token)
        except pydantic.ValidationError as exc:
            raise InvalidTokenPayload() from exc

        lexicon_permissions = {}
        if payload.scope and "lexica" in payload.scope:
            lexicon_permissions = payload.scope["lexica"]
        return User(payload.sub, lexicon_permissions, payload.levels)

    def authorize(
        self,
        level: value_objects.PermissionLevel,
        user: User,
        resource_ids: List[str],
    ):
        return not any(
            self.is_resource_protected.query(resource_id, level)
            and (not user or not user.has_enough_permissions(resource_id, level))
            for resource_id in resource_ids
        )



from karp.webapp.dependencies.lex_deps import (
    get_entry_diff,
    get_entry_history,
    get_history,
    get_lex_uc,
    get_resources_read_repo,
)
from karp.webapp.dependencies.auth_deps import (
    get_auth_service,
    get_user,
    get_user_optional,
    get_resource_permissions,
)
from karp.webapp.dependencies.fastapi_injector import inject_from_req
